from typing import Optional, Tuple

from click.exceptions import ClickException

from anyscale.cli_logger import BlockLogger
from anyscale.client.openapi_client.api.default_api import DefaultApi
from anyscale.cloud import get_cloud_id_and_name


def get_organization_default_cloud(api_client: DefaultApi) -> Optional[str]:
    """Return default cloud name for organization if it exists and
        if user has correct permissions for it.

        Returns:
            Name of default cloud name for organization if it exists and
            if user has correct permissions for it.
        """
    user = api_client.get_user_info_api_v2_userinfo_get().result
    organization = user.organizations[0]  # Each user only has one org
    if organization.default_cloud_id:
        try:
            # Check permissions
            _, cloud_name = get_cloud_id_and_name(
                api_client, cloud_id=organization.default_cloud_id
            )
            return str(cloud_name)
        except Exception:  # noqa: BLE001
            return None
    return None


def get_default_cloud(
    api_client: DefaultApi, cloud_name: Optional[str]
) -> Tuple[str, str]:
    """Returns the cloud id from cloud name.
    If cloud name is not provided, returns the default cloud name if exists in organization.
    If default cloud name does not exist returns last used cloud.
    """

    if cloud_name is None:
        default_cloud_name = get_organization_default_cloud(api_client)
        if default_cloud_name:
            cloud_name = default_cloud_name
    return get_cloud_id_and_name(api_client, cloud_name=cloud_name)


def verify_anyscale_access(
    api_client: DefaultApi, cloud_id: str, logger: BlockLogger
) -> bool:
    try:
        api_client.verify_access_api_v2_cloudsverify_access_cloud_id_get(cloud_id)
        return True
    except ClickException as e:
        logger.error(
            f"Anyscale's control plane is unable to access resources on your cloud provider.\n{e}"
        )
        return False



import click


# TODO: type the return value
def try_import_gcp_secretmanager():
    try:
        from google.cloud import secretmanager

        return secretmanager
    except ImportError:
        raise click.ClickException(
            "pip package `google-cloud-secret-manager` is not installed locally on this machine but required "
            "for the command. Please install with `pip install 'anyscale[gcp]'`."
        )


def try_import_gcp_discovery():
    try:
        from googleapiclient import discovery

        return discovery
    except ImportError:
        raise click.ClickException(
            "pip package `google-api-python-client` is not installed locally on this machine but required "
            "for the command. Please install with `pip install 'anyscale[gcp]'`."
        )


def try_import_gcp_utils():
    try:
        from anyscale.utils import gcp_utils

        return gcp_utils
    except ImportError as e:
        raise click.ClickException(
            f"Import error occurred: {e}. Please install GCP-related packages with `pip install 'anyscale[gcp]'`."
        )


def try_import_gcp_verify_lib():
    try:
        import anyscale.gcp_verification as verify_lib

        return verify_lib
    except ImportError as e:
        raise click.ClickException(
            f"Import error occurred: {e}. Please install GCP-related packages with `pip install 'anyscale[gcp]'`."
        )



# AUTOGENERATED - modify shared_anyscale_util in root directory to make changes
import asyncio
import logging
from typing import Awaitable, List, Optional, TypeVar, Union


_T = TypeVar("_T")
logger = logging.getLogger(__file__)


async def gather_in_batches(
    n: int,
    *tasks: Awaitable[_T],
    per_task_timeout_secs: Optional[float] = None,
    return_exceptions: bool = False,
    log_exceptions: bool = False,
) -> List[Union[_T, BaseException]]:
    # n cannot be less than 1
    n = max(n, 1)
    semaphore = asyncio.Semaphore(n)

    async def sem_task(task: Awaitable[_T]) -> _T:
        try:
            async with semaphore:
                return await asyncio.wait_for(task, per_task_timeout_secs)
        except Exception:
            if log_exceptions:
                logger.exception("Error while gathering task")
            raise  # re-raise exception

    return await asyncio.gather(
        *(sem_task(task) for task in tasks), return_exceptions=return_exceptions
    )



from typing import Optional

from anyscale.sdk.anyscale_client.api.default_api import DefaultApi as BaseApi
from anyscale.sdk.anyscale_client.models.cloud import Cloud
from anyscale.sdk.anyscale_client.models.cloud_version import CloudVersion


def _is_v2_cloud(base_api: BaseApi, cloud_id: str) -> bool:
    cloud: Cloud = base_api.get_cloud(cloud_id).result
    cloud_version: Optional[CloudVersion] = cloud.version
    return not cloud_version or cloud_version == CloudVersion.V2



import json
import os
import shutil
import tempfile
from typing import Optional
from unittest.mock import MagicMock, patch

from anyscale import component_activity_util


class StatusFile:
    is_active: str
    reason: str
    timestamp: float
    last_activity_timestamp: float


class TestComponentActivityEnv:
    def __init__(self, idle_termination_dir: Optional[str] = None):
        if idle_termination_dir:
            self.idle_termination_dir = idle_termination_dir
            component_activity_util.IDLE_TERMINATION_DIR = idle_termination_dir

    def write_status_file(self, file_name: str, content):
        with open(f"{self.idle_termination_dir}/{file_name}.json", "w") as f:
            f.write(json.dumps(content))

    def __del__(self):
        if self.idle_termination_dir and os.path.exists(self.idle_termination_dir):
            shutil.rmtree(self.idle_termination_dir)


@patch("time.time", MagicMock(return_value=12345.6789))
def test_valid_and_empty_status():
    env = TestComponentActivityEnv(tempfile.mkdtemp())

    # Set a valid workspace status
    workspace_last_activity_timestamp = 9876.54321
    content = {
        "last_activity_timestamp": workspace_last_activity_timestamp,
    }
    env.write_status_file("workspace", content)

    # Do not set a web_terminal status

    result = component_activity_util.env_hook()
    assert result, "Env hook return none"
    assert result["workspace"], "workspace status should not be None"
    assert result["web_terminal"], "web_terminal status should not be None"
    workspace_result = result["workspace"]

    # Verify the workspace status matches input
    assert workspace_result["is_active"] == "INACTIVE"
    assert workspace_result["reason"] == "workspace last snapshot"
    assert workspace_result["timestamp"] == 12345.6789
    assert workspace_result["last_activity_at"] == workspace_last_activity_timestamp

    # Verify the web terminal status is empty
    web_terminal_result = result["web_terminal"]
    assert web_terminal_result["is_active"] == "ERROR"
    assert web_terminal_result["reason"] == "status file does not exist"
    assert web_terminal_result["timestamp"] == 12345.6789
    assert not web_terminal_result.get("last_activity_at")


@patch("time.time", MagicMock(return_value=12345.6789))
def test_error_status():
    env = TestComponentActivityEnv(tempfile.mkdtemp())
    # Valid status for workspace
    content = {
        "error": "custom error message",
    }
    env.write_status_file("workspace", content)

    # Invalid status for the web terminal
    env.write_status_file("web_terminal", {"invalid_property": 1234})

    result = component_activity_util.env_hook()
    assert result, "Env hook return none"
    assert result["workspace"], "workspace status should not be None"
    assert result["web_terminal"], "web_terminal status should not be None"
    workspace_result = result["workspace"]

    # Verify the workspace status is reported correctly
    assert workspace_result["is_active"] == "ERROR"
    assert workspace_result["reason"] == "custom error message"
    assert workspace_result["timestamp"] == 12345.6789
    assert not workspace_result.get("last_activity_at")

    # Verify an empty web terminal status
    web_terminal_result = result["web_terminal"]
    assert web_terminal_result["is_active"] == "ERROR"
    assert web_terminal_result["reason"] == "error parsing the status file"
    assert web_terminal_result["timestamp"] == 12345.6789
    assert not web_terminal_result.get("last_activity_at")



import asyncio
from contextlib import contextmanager
from http import HTTPStatus
import json
from typing import Generator, Optional

from aiohttp import ClientConnectionError, ClientResponseError
from anyscale_client.exceptions import ApiException
from click import ClickException


class ExpectedLogRetrievalError(Exception):
    pass


class UnsupportedLogRetrievalMethodError(ExpectedLogRetrievalError):
    pass


class LogRetrievalTimeoutError(ExpectedLogRetrievalError):
    pass


class LogNotFoundOnActiveClusterError(ExpectedLogRetrievalError):
    pass


@contextmanager
def wrap_as_unsupported_log_retrieval_method_error() -> Generator[None, None, None]:
    try:
        yield
    except ApiException as e:
        raise UnsupportedLogRetrievalMethodError(
            _get_exception_message(e)
        ) if e.status == HTTPStatus.BAD_REQUEST else e


def _get_exception_message(e: ApiException) -> Optional[str]:
    return json.loads(e.body).get("error", {}).get("detail")


@contextmanager
def wrap_job_run_log_not_retrievable_on_active_cluster_error(
    job_run_id: str, raise_connection_issue_as_cli_error: bool = False
) -> Generator[None, None, None]:
    try:
        yield
    except (ClientConnectionError, asyncio.TimeoutError):
        error_msg = (
            f'Retrieving logs while the job is running is only supported if the Anyscale client (SDK / CLI) can reach the cluster running the job "{job_run_id}".\n'
            "Check that you are calling this from the same VPN as the cluster, or if the cluster is unreachable because of high load."
        )
        if raise_connection_issue_as_cli_error:
            raise ClickException(error_msg)
        else:
            raise UnsupportedLogRetrievalMethodError(error_msg)
    except ClientResponseError as e:
        # Will return 404 if job run (Ray job) is not found
        # https://github.com/ray-project/ray/blob/6334576223dd916469fcdac185752b80e0c3e416/dashboard/modules/job/job_head.py#L485
        raise LogNotFoundOnActiveClusterError(
            f'Job run "{job_run_id}" is not a part of the current active session of the cluster. '
            "The cluster may have restarted since the job ran."
        ) if e.status == HTTPStatus.NOT_FOUND else e



# coding: utf-8

"""
    Managed Ray API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six

from openapi_client.configuration import Configuration


class CreateCloudWithCloudResource(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'name': 'str',
        'provider': 'CloudProviders',
        'region': 'str',
        'credentials': 'str',
        'config': 'CloudConfig',
        'is_k8s': 'bool',
        'is_aioa': 'bool',
        'availability_zones': 'list[str]',
        'is_bring_your_own_resource': 'bool',
        'is_private_cloud': 'bool',
        'cluster_management_stack_version': 'ClusterManagementStackVersions',
        'cloud_resource_to_create': 'CreateCloudResource'
    }

    attribute_map = {
        'name': 'name',
        'provider': 'provider',
        'region': 'region',
        'credentials': 'credentials',
        'config': 'config',
        'is_k8s': 'is_k8s',
        'is_aioa': 'is_aioa',
        'availability_zones': 'availability_zones',
        'is_bring_your_own_resource': 'is_bring_your_own_resource',
        'is_private_cloud': 'is_private_cloud',
        'cluster_management_stack_version': 'cluster_management_stack_version',
        'cloud_resource_to_create': 'cloud_resource_to_create'
    }

    def __init__(self, name=None, provider=None, region=None, credentials=None, config=None, is_k8s=False, is_aioa=False, availability_zones=None, is_bring_your_own_resource=None, is_private_cloud=False, cluster_management_stack_version=None, cloud_resource_to_create=None, local_vars_configuration=None):  # noqa: E501
        """CreateCloudWithCloudResource - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration()
        self.local_vars_configuration = local_vars_configuration

        self._name = None
        self._provider = None
        self._region = None
        self._credentials = None
        self._config = None
        self._is_k8s = None
        self._is_aioa = None
        self._availability_zones = None
        self._is_bring_your_own_resource = None
        self._is_private_cloud = None
        self._cluster_management_stack_version = None
        self._cloud_resource_to_create = None
        self.discriminator = None

        self.name = name
        self.provider = provider
        self.region = region
        self.credentials = credentials
        if config is not None:
            self.config = config
        if is_k8s is not None:
            self.is_k8s = is_k8s
        if is_aioa is not None:
            self.is_aioa = is_aioa
        if availability_zones is not None:
            self.availability_zones = availability_zones
        if is_bring_your_own_resource is not None:
            self.is_bring_your_own_resource = is_bring_your_own_resource
        if is_private_cloud is not None:
            self.is_private_cloud = is_private_cloud
        if cluster_management_stack_version is not None:
            self.cluster_management_stack_version = cluster_management_stack_version
        if cloud_resource_to_create is not None:
            self.cloud_resource_to_create = cloud_resource_to_create

    @property
    def name(self):
        """Gets the name of this CreateCloudWithCloudResource.  # noqa: E501

        Name of this Cloud.  # noqa: E501

        :return: The name of this CreateCloudWithCloudResource.  # noqa: E501
        :rtype: str
        """
        return self._name

    @name.setter
    def name(self, name):
        """Sets the name of this CreateCloudWithCloudResource.

        Name of this Cloud.  # noqa: E501

        :param name: The name of this CreateCloudWithCloudResource.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and name is None:  # noqa: E501
            raise ValueError("Invalid value for `name`, must not be `None`")  # noqa: E501

        self._name = name

    @property
    def provider(self):
        """Gets the provider of this CreateCloudWithCloudResource.  # noqa: E501

        Provider of this Cloud (e.g. AWS).  # noqa: E501

        :return: The provider of this CreateCloudWithCloudResource.  # noqa: E501
        :rtype: CloudProviders
        """
        return self._provider

    @provider.setter
    def provider(self, provider):
        """Sets the provider of this CreateCloudWithCloudResource.

        Provider of this Cloud (e.g. AWS).  # noqa: E501

        :param provider: The provider of this CreateCloudWithCloudResource.  # noqa: E501
        :type: CloudProviders
        """
        if self.local_vars_configuration.client_side_validation and provider is None:  # noqa: E501
            raise ValueError("Invalid value for `provider`, must not be `None`")  # noqa: E501

        self._provider = provider

    @property
    def region(self):
        """Gets the region of this CreateCloudWithCloudResource.  # noqa: E501

        Region this Cloud is operating in. This value needs to be supported by this Cloud's provider. (e.g. us-west-2)  # noqa: E501

        :return: The region of this CreateCloudWithCloudResource.  # noqa: E501
        :rtype: str
        """
        return self._region

    @region.setter
    def region(self, region):
        """Sets the region of this CreateCloudWithCloudResource.

        Region this Cloud is operating in. This value needs to be supported by this Cloud's provider. (e.g. us-west-2)  # noqa: E501

        :param region: The region of this CreateCloudWithCloudResource.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and region is None:  # noqa: E501
            raise ValueError("Invalid value for `region`, must not be `None`")  # noqa: E501

        self._region = region

    @property
    def credentials(self):
        """Gets the credentials of this CreateCloudWithCloudResource.  # noqa: E501

        Credentials needed to interact with this Cloud.  # noqa: E501

        :return: The credentials of this CreateCloudWithCloudResource.  # noqa: E501
        :rtype: str
        """
        return self._credentials

    @credentials.setter
    def credentials(self, credentials):
        """Sets the credentials of this CreateCloudWithCloudResource.

        Credentials needed to interact with this Cloud.  # noqa: E501

        :param credentials: The credentials of this CreateCloudWithCloudResource.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and credentials is None:  # noqa: E501
            raise ValueError("Invalid value for `credentials`, must not be `None`")  # noqa: E501

        self._credentials = credentials

    @property
    def config(self):
        """Gets the config of this CreateCloudWithCloudResource.  # noqa: E501

        Additional configurable properties of this Cloud.  # noqa: E501

        :return: The config of this CreateCloudWithCloudResource.  # noqa: E501
        :rtype: CloudConfig
        """
        return self._config

    @config.setter
    def config(self, config):
        """Sets the config of this CreateCloudWithCloudResource.

        Additional configurable properties of this Cloud.  # noqa: E501

        :param config: The config of this CreateCloudWithCloudResource.  # noqa: E501
        :type: CloudConfig
        """

        self._config = config

    @property
    def is_k8s(self):
        """Gets the is_k8s of this CreateCloudWithCloudResource.  # noqa: E501

        Whether this cloud is managed via Kubernetes.  # noqa: E501

        :return: The is_k8s of this CreateCloudWithCloudResource.  # noqa: E501
        :rtype: bool
        """
        return self._is_k8s

    @is_k8s.setter
    def is_k8s(self, is_k8s):
        """Sets the is_k8s of this CreateCloudWithCloudResource.

        Whether this cloud is managed via Kubernetes.  # noqa: E501

        :param is_k8s: The is_k8s of this CreateCloudWithCloudResource.  # noqa: E501
        :type: bool
        """

        self._is_k8s = is_k8s

    @property
    def is_aioa(self):
        """Gets the is_aioa of this CreateCloudWithCloudResource.  # noqa: E501

        Whether this cloud is an AIOA cloud.  # noqa: E501

        :return: The is_aioa of this CreateCloudWithCloudResource.  # noqa: E501
        :rtype: bool
        """
        return self._is_aioa

    @is_aioa.setter
    def is_aioa(self, is_aioa):
        """Sets the is_aioa of this CreateCloudWithCloudResource.

        Whether this cloud is an AIOA cloud.  # noqa: E501

        :param is_aioa: The is_aioa of this CreateCloudWithCloudResource.  # noqa: E501
        :type: bool
        """

        self._is_aioa = is_aioa

    @property
    def availability_zones(self):
        """Gets the availability_zones of this CreateCloudWithCloudResource.  # noqa: E501

        The availability zones that instances of this cloud are allowed to be launched in.  # noqa: E501

        :return: The availability_zones of this CreateCloudWithCloudResource.  # noqa: E501
        :rtype: list[str]
        """
        return self._availability_zones

    @availability_zones.setter
    def availability_zones(self, availability_zones):
        """Sets the availability_zones of this CreateCloudWithCloudResource.

        The availability zones that instances of this cloud are allowed to be launched in.  # noqa: E501

        :param availability_zones: The availability_zones of this CreateCloudWithCloudResource.  # noqa: E501
        :type: list[str]
        """

        self._availability_zones = availability_zones

    @property
    def is_bring_your_own_resource(self):
        """Gets the is_bring_your_own_resource of this CreateCloudWithCloudResource.  # noqa: E501

        Whether the resources of this cloud are provided by the customer.  # noqa: E501

        :return: The is_bring_your_own_resource of this CreateCloudWithCloudResource.  # noqa: E501
        :rtype: bool
        """
        return self._is_bring_your_own_resource

    @is_bring_your_own_resource.setter
    def is_bring_your_own_resource(self, is_bring_your_own_resource):
        """Sets the is_bring_your_own_resource of this CreateCloudWithCloudResource.

        Whether the resources of this cloud are provided by the customer.  # noqa: E501

        :param is_bring_your_own_resource: The is_bring_your_own_resource of this CreateCloudWithCloudResource.  # noqa: E501
        :type: bool
        """

        self._is_bring_your_own_resource = is_bring_your_own_resource

    @property
    def is_private_cloud(self):
        """Gets the is_private_cloud of this CreateCloudWithCloudResource.  # noqa: E501

        Whether this cloud is a private cloud.  # noqa: E501

        :return: The is_private_cloud of this CreateCloudWithCloudResource.  # noqa: E501
        :rtype: bool
        """
        return self._is_private_cloud

    @is_private_cloud.setter
    def is_private_cloud(self, is_private_cloud):
        """Sets the is_private_cloud of this CreateCloudWithCloudResource.

        Whether this cloud is a private cloud.  # noqa: E501

        :param is_private_cloud: The is_private_cloud of this CreateCloudWithCloudResource.  # noqa: E501
        :type: bool
        """

        self._is_private_cloud = is_private_cloud

    @property
    def cluster_management_stack_version(self):
        """Gets the cluster_management_stack_version of this CreateCloudWithCloudResource.  # noqa: E501

        The cluster management stack version of the cloud.  # noqa: E501

        :return: The cluster_management_stack_version of this CreateCloudWithCloudResource.  # noqa: E501
        :rtype: ClusterManagementStackVersions
        """
        return self._cluster_management_stack_version

    @cluster_management_stack_version.setter
    def cluster_management_stack_version(self, cluster_management_stack_version):
        """Sets the cluster_management_stack_version of this CreateCloudWithCloudResource.

        The cluster management stack version of the cloud.  # noqa: E501

        :param cluster_management_stack_version: The cluster_management_stack_version of this CreateCloudWithCloudResource.  # noqa: E501
        :type: ClusterManagementStackVersions
        """

        self._cluster_management_stack_version = cluster_management_stack_version

    @property
    def cloud_resource_to_create(self):
        """Gets the cloud_resource_to_create of this CreateCloudWithCloudResource.  # noqa: E501

        The cloud resource of this cloud.  # noqa: E501

        :return: The cloud_resource_to_create of this CreateCloudWithCloudResource.  # noqa: E501
        :rtype: CreateCloudResource
        """
        return self._cloud_resource_to_create

    @cloud_resource_to_create.setter
    def cloud_resource_to_create(self, cloud_resource_to_create):
        """Sets the cloud_resource_to_create of this CreateCloudWithCloudResource.

        The cloud resource of this cloud.  # noqa: E501

        :param cloud_resource_to_create: The cloud_resource_to_create of this CreateCloudWithCloudResource.  # noqa: E501
        :type: CreateCloudResource
        """

        self._cloud_resource_to_create = cloud_resource_to_create

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, CreateCloudWithCloudResource):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, CreateCloudWithCloudResource):
            return True

        return self.to_dict() != other.to_dict()



import asyncio
import re
from typing import Any, Callable, List

import aiohttp
from aiohttp import WSMsgType

from anyscale.shared_anyscale_utils.utils.asyncio import gather_in_batches


CLUSTER_CONNECT_TIMEOUT = 30


async def _download_logs_concurrently(
    log_chunk_urls: List[str], parallelism: int
) -> str:
    logs_across_chunks: List[str] = await gather_in_batches(  # type: ignore
        parallelism, *[_download_log_from_s3_url(url) for url in log_chunk_urls]
    )
    logs_across_chunks = [log.strip() for log in logs_across_chunks]
    return "\n".join(logs_across_chunks)


async def _download_log_from_ray_json_response(url: str) -> str:
    async with aiohttp.ClientSession() as session:
        response = await asyncio.wait_for(
            session.get(url), timeout=CLUSTER_CONNECT_TIMEOUT
        )
        logs: str = (await response.json()).get("logs", "")
        return logs


async def _download_log_from_s3_url(url: str) -> str:
    # Note that the URL is presigned, so no token needs to be passed in the request
    async with aiohttp.ClientSession() as session:
        async with session.get(url) as response:
            return await response.text()


# For job logs, the implementation is same as
# https://github.com/ray-project/ray/blob/6142f52ada995b7123ea7347a70613ca7b924e0c/dashboard/modules/job/sdk.py#L426
# However Ray is only part of [all] install, so we replicate the functionality here.
async def _stream_log_from_ray_websocket(
    url: str, log_callback: Callable[[str], Any], remove_escape_chars: bool = True
) -> None:
    async with aiohttp.ClientSession(
        timeout=aiohttp.ClientTimeout(total=CLUSTER_CONNECT_TIMEOUT)
    ) as session:
        async with session.ws_connect(url) as ws:
            while True:
                msg = await ws.receive()
                if msg.type == WSMsgType.TEXT:
                    logs = msg.data.strip()
                    if remove_escape_chars:
                        logs = _remove_ansi_escape_sequences(logs)
                    log_callback(logs)
                elif msg.type == WSMsgType.CLOSED:
                    break
                elif msg.type == WSMsgType.ERROR:
                    pass


def _remove_ansi_escape_sequences(s: str) -> str:
    # Required as the log may contain ANSI escape sequeneces (e.g. for coloring in the terminal)
    # Regex pattern from https://stackoverflow.com/a/14693789
    ansi_escape = re.compile(r"\x1B(?:[@-Z\\-_]|\[[0-?]*[ -/]*[@-~])")
    return ansi_escape.sub("", s)



from contextlib import contextmanager
import os
import sys
import time
from typing import Any, Optional

import click
import colorama  # type: ignore
from halo import Halo


_process_start_time = time.time()


class BlockLogger:
    """
    Logger class for the CLI. Supports formatting in blocks if `block_label` is provided to
    the methods. Also supports anyscale.connect style logging if `block_label` is not
    provided.
    """

    def __init__(
        self,
        log_output: bool = True,
        t0: float = _process_start_time,
        spinner_manager: Any = None,
    ) -> None:
        self.t0 = t0
        # Flag to disable all terminal output from CLILogger (useful for SDK)
        self.log_output = log_output
        self.current_block: Optional[str] = None
        self.spinner_manager = spinner_manager

    def open_block(
        self, block_label: str, block_title: Optional[str] = None, auto_close=False
    ) -> None:
        """
        Prints block title from the given block_label and sets self.current_block. "Output"
        is a generic block that does not need to follow the standard convention.

        if auto_close is set, automatically closes the block before opening a new one
        """
        if not self.log_output:
            return
        assert (
            auto_close or self.current_block is None or self.current_block == "Output"
        ), f"Block {self.current_block} is already open. Please close before opening block {block_label}."

        if auto_close:
            self.close_block()
        self.current_block = block_label
        print(
            f"{colorama.Style.BRIGHT}{colorama.Fore.CYAN}{block_title if block_title else block_label}{colorama.Style.RESET_ALL}",
            file=sys.stderr,
        )

    def close_block(self, block_label: Optional[str] = None) -> None:
        """ Closes the current block
        If a label is specified, it must match the current open block label.
        raise an AssertionError if we try to close a different block

        Prints newline so there is separation before next block is opened.
        """
        if not self.log_output:
            return
        if block_label:
            assert (
                self.current_block == block_label
            ), f"Attempting to close block {block_label}, but block {self.current_block} is currently open."
        self.current_block = None
        print(file=sys.stderr)

    @staticmethod
    def highlight(word: str) -> str:
        return f"{colorama.Style.BRIGHT}{colorama.Fore.MAGENTA}{word}{colorama.Style.RESET_ALL}"

    def zero_time(self) -> None:
        self.t0 = time.time()

    def info(self, *msg: str, block_label: Optional[str] = None) -> None:
        if not self.log_output:
            return
        if block_label:
            # Check block_label if provided.
            assert (
                self.current_block == block_label
            ), f"Attempting to log to block {block_label}, but block {self.current_block} is currently open."
            print(
                *msg, file=sys.stderr,
            )
        else:
            print(
                "{}{}(anyscale +{}){} ".format(
                    colorama.Style.BRIGHT,
                    colorama.Fore.CYAN,
                    self._time_string(),
                    colorama.Style.RESET_ALL,
                ),
                end="",
                file=sys.stderr,
            )
            print(
                *msg, file=sys.stderr,
            )

    def debug(self, *msg: str) -> None:
        if not self.log_output:
            return
        if os.environ.get("ANYSCALE_DEBUG") == "1":
            print(
                "{}{}(anyscale +{}){} ".format(
                    colorama.Style.DIM,
                    colorama.Fore.CYAN,
                    self._time_string(),
                    colorama.Style.RESET_ALL,
                ),
                end="",
            )
            print(*msg)

    def warning(self, *msg: str) -> None:
        if not self.log_output:
            return
        print(
            "{}{}[Warning]{} ".format(
                colorama.Style.NORMAL, colorama.Fore.YELLOW, colorama.Style.RESET_ALL,
            ),
            end="",
            file=sys.stderr,
        )
        print(*msg, file=sys.stderr)

    def _time_string(self) -> str:
        delta = time.time() - self.t0
        hours = 0
        minutes = 0
        while delta > 3600:
            hours += 1
            delta -= 3600
        while delta > 60:
            minutes += 1
            delta -= 60
        output = ""
        if hours:
            output += f"{hours}h"
        if minutes:
            output += f"{minutes}m"
        output += f"{round(delta, 1)}s"
        return output

    def error(self, *msg: str) -> None:
        print(
            "{}{}(anyscale +{}){} ".format(
                colorama.Style.BRIGHT,
                colorama.Fore.RED,
                self._time_string(),
                colorama.Style.RESET_ALL,
            ),
            end="",
            file=sys.stderr,
        )
        print(*msg, file=sys.stderr)

    def confirm_missing_permission(self, error_str: str):
        self.warning(error_str)
        if self.spinner_manager:
            self.spinner_manager.stop()

        click.confirm(
            "If the IAM role has these required permissions granted some other ways"
            "or if you would like to continue without these permissions, press 'y' to continue with verification or press 'N' to abort.",
            abort=True,
        )

        if self.spinner_manager:
            self.spinner_manager.start()


class LogsLogger(BlockLogger):
    """ This logger is used to print customer logs to STDOUT with no decoration
    """

    def log(self, msg: str):
        print(msg)

    def is_interactive_cli_enabled(self) -> bool:
        """Check if shell is interactive
        """

        default_env_var = "1" if sys.stderr.isatty() else "0"
        return os.environ.get("ANYSCALE_CLI_INTERACTIVE_UX", default_env_var) == "1"

    @contextmanager
    def spinner(self, msg: str):
        """ Create a spinner with the starting text.
        To update the text next to the spinner, set `spinner.text = "new_text"`
        The spinner will be stopped when the context is exited.
        Spinner is only enabled for interactive shell
        """

        enable_spinner = self.is_interactive_cli_enabled()

        with Halo(
            text=msg, spinner="dots", stream=sys.stderr, enabled=enable_spinner
        ) as spinner:
            yield spinner



# coding: utf-8

"""
    Managed Ray API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six

from openapi_client.configuration import Configuration


class UpdateCloudWithCloudResourceGCP(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'availability_zones': 'list[str]',
        'state': 'CloudState',
        'cloud_resource_to_update': 'CreateCloudResourceGCP'
    }

    attribute_map = {
        'availability_zones': 'availability_zones',
        'state': 'state',
        'cloud_resource_to_update': 'cloud_resource_to_update'
    }

    def __init__(self, availability_zones=None, state=None, cloud_resource_to_update=None, local_vars_configuration=None):  # noqa: E501
        """UpdateCloudWithCloudResourceGCP - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration()
        self.local_vars_configuration = local_vars_configuration

        self._availability_zones = None
        self._state = None
        self._cloud_resource_to_update = None
        self.discriminator = None

        if availability_zones is not None:
            self.availability_zones = availability_zones
        if state is not None:
            self.state = state
        self.cloud_resource_to_update = cloud_resource_to_update

    @property
    def availability_zones(self):
        """Gets the availability_zones of this UpdateCloudWithCloudResourceGCP.  # noqa: E501

        The availability zones that instances of this cloud are allowed to be launched in.  # noqa: E501

        :return: The availability_zones of this UpdateCloudWithCloudResourceGCP.  # noqa: E501
        :rtype: list[str]
        """
        return self._availability_zones

    @availability_zones.setter
    def availability_zones(self, availability_zones):
        """Sets the availability_zones of this UpdateCloudWithCloudResourceGCP.

        The availability zones that instances of this cloud are allowed to be launched in.  # noqa: E501

        :param availability_zones: The availability_zones of this UpdateCloudWithCloudResourceGCP.  # noqa: E501
        :type: list[str]
        """

        self._availability_zones = availability_zones

    @property
    def state(self):
        """Gets the state of this UpdateCloudWithCloudResourceGCP.  # noqa: E501

        The state of this cloud.  # noqa: E501

        :return: The state of this UpdateCloudWithCloudResourceGCP.  # noqa: E501
        :rtype: CloudState
        """
        return self._state

    @state.setter
    def state(self, state):
        """Sets the state of this UpdateCloudWithCloudResourceGCP.

        The state of this cloud.  # noqa: E501

        :param state: The state of this UpdateCloudWithCloudResourceGCP.  # noqa: E501
        :type: CloudState
        """

        self._state = state

    @property
    def cloud_resource_to_update(self):
        """Gets the cloud_resource_to_update of this UpdateCloudWithCloudResourceGCP.  # noqa: E501

        The cloud resource of this GCP cloud.  # noqa: E501

        :return: The cloud_resource_to_update of this UpdateCloudWithCloudResourceGCP.  # noqa: E501
        :rtype: CreateCloudResourceGCP
        """
        return self._cloud_resource_to_update

    @cloud_resource_to_update.setter
    def cloud_resource_to_update(self, cloud_resource_to_update):
        """Sets the cloud_resource_to_update of this UpdateCloudWithCloudResourceGCP.

        The cloud resource of this GCP cloud.  # noqa: E501

        :param cloud_resource_to_update: The cloud_resource_to_update of this UpdateCloudWithCloudResourceGCP.  # noqa: E501
        :type: CreateCloudResourceGCP
        """
        if self.local_vars_configuration.client_side_validation and cloud_resource_to_update is None:  # noqa: E501
            raise ValueError("Invalid value for `cloud_resource_to_update`, must not be `None`")  # noqa: E501

        self._cloud_resource_to_update = cloud_resource_to_update

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, UpdateCloudWithCloudResourceGCP):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, UpdateCloudWithCloudResourceGCP):
            return True

        return self.to_dict() != other.to_dict()



"""
Fetches required data and formats output for `anyscale list` commands.
"""

import json
import time
from typing import Any, Dict, List, Optional, Tuple

import tabulate

from anyscale.cli_logger import BlockLogger
from anyscale.cloud import get_cloud_json_from_id
from anyscale.cluster_config import get_cluster_config
from anyscale.controllers.base_controller import BaseController
from anyscale.formatters import clouds_formatter
from anyscale.project import (
    get_proj_name_from_id,
    get_project_id,
    get_project_session,
    get_project_sessions,
    load_project_or_throw,
)
from anyscale.util import get_endpoint, humanize_timestamp
from anyscale.utils.imports.all import try_import_ray


class ListController(BaseController):
    def __init__(
        self, log: Optional[BlockLogger] = None, initialize_auth_api_client: bool = True
    ):
        if log is None:
            log = BlockLogger()

        super().__init__(initialize_auth_api_client=initialize_auth_api_client)
        self.log = log
        self.log.open_block("Output")

    def list_clouds(self, json_format: bool) -> str:
        clouds = self.api_client.list_clouds_api_v2_clouds_get().results
        output = clouds_formatter.format_clouds_output(
            clouds=clouds, json_format=json_format
        )

        return str(output)

    # TODO (jbai): This requires additional refactoring to decouple network fetching from formatting logic
    def list_projects(self, json_format: bool) -> str:
        resp = self.api_client.list_projects_api_v2_projects_get()
        projects = resp.results
        project_table = []
        project_json = []

        for project in projects:
            if json_format:
                session_list = self.api_client.list_sessions_api_v2_sessions_get(
                    project_id=project.id
                ).results
                session_record = []
                for session in session_list:
                    status = self.get_session_status_and_commands(session=session)[0]
                    session_record.append({"name": session.name, "status": status})
                project_json.append(
                    {
                        "name": project.name,
                        "sessions": session_record,
                        "url": "{}".format(get_endpoint(f"/projects/{project.id}")),
                    }
                )
            else:
                project_table.append(
                    [
                        project.name,
                        get_endpoint(f"/projects/{project.id}"),
                        project.description,
                    ]
                )

        if json_format:
            return json.dumps(project_json)
        else:
            table = tabulate.tabulate(
                project_table, headers=["NAME", "URL", "DESCRIPTION"], tablefmt="plain",
            )
            return f"Projects:\n{table}"

    # TODO (jbai): This requires additional refactoring to decouple network fetching from formatting logic
    def list_sessions(  # noqa: PLR0912
        self, name: Optional[str], show_all: bool, json_format: bool
    ) -> str:
        project_definition = load_project_or_throw()
        project_id = get_project_id(project_definition.root)

        sessions = self.api_client.list_sessions_api_v2_sessions_get(
            project_id=project_id, name=name, active_only=not show_all
        ).results

        if json_format:
            output = self.session_list_json(sessions=sessions, project_id=project_id)
            return output

        active_project_title = f"Active project: {project_definition.root}"
        if name is None:
            table = []
            for session in sessions:
                created_at = humanize_timestamp(session.created_at)

                idle_timeout_minutes = (
                    session.idle_timeout
                )  # the constant idle_timeout specified by the user
                if idle_timeout_minutes >= 0:
                    idle_time_remaining_seconds = (
                        session.idle_time_remaining_seconds
                    )  # the remaining time to auto-suspend in seconds

                    if idle_time_remaining_seconds is not None:
                        if idle_time_remaining_seconds < 60:
                            # If we have LESS than a minute left till auto-suspend
                            idle_time_remaining_message = " ({} s left)".format(
                                idle_time_remaining_seconds
                            )
                        else:
                            # If we have MORE than a minute left till auto-suspend
                            idle_time_remaining_message = " ({} min left)".format(
                                int(round(idle_time_remaining_seconds / 60))
                            )
                    else:
                        # Even if the idle_timeout_minutes >= 0, the
                        # idle_time_remaining_seconds might still be None if
                        # the session is not in the "Running" state.
                        idle_time_remaining_message = ""

                    idle_timeout_message = "Enabled{}".format(
                        idle_time_remaining_message
                    )
                else:
                    idle_timeout_message = "Disabled"

                record = [
                    session.name,
                    " {}".format(
                        self.get_session_status_and_commands(session=session)[0]
                    ),
                    created_at,
                    idle_timeout_message,
                ]
                if show_all:
                    table.append(
                        [
                            " Y"
                            if session.state
                            in [
                                "Running",
                                "Updating",
                                "UpdatingErrored",
                                "StartingUp",
                                "StartupErrored",
                            ]
                            else " N"
                        ]
                        + record
                    )
                else:
                    table.append(record)
            if not show_all:
                table_string = tabulate.tabulate(
                    table,
                    headers=["SESSION", "STATUS", "CREATED", "AUTO-SUSPEND"],
                    tablefmt="plain",
                )
            else:
                table_string = tabulate.tabulate(
                    table,
                    headers=["ACTIVE", "SESSION", "STATUS", "CREATED", "AUTO-SUSPEND"],
                    tablefmt="plain",
                )

            return f"{active_project_title}\n{table_string}"
        else:
            sessions = [session for session in sessions if session.name == name]
            session_output_strings = []
            for session in sessions:
                resp = self.api_client.describe_session_api_v2_sessions_session_id_describe_get(
                    session_id=session.id
                )

                snapshot_table = []
                for applied_snapshot in resp.result.applied_snapshots:
                    snapshot_id = applied_snapshot.snapshot_id
                    created_at = humanize_timestamp(applied_snapshot.created_at)
                    snapshot_table.append([snapshot_id, created_at])
                session_output_strings.append(
                    tabulate.tabulate(
                        snapshot_table,
                        headers=[f"SNAPSHOT applied to {session.name}", "APPLIED",],
                        tablefmt="plain",
                    )
                )

                command_table = []
                for command in resp.results.commands:
                    created_at = humanize_timestamp(command.created_at)
                    command_table.append(
                        [
                            " ".join(
                                [command.name]
                                + [
                                    f"{key}={val}"
                                    for key, val in command.params.items()
                                ]
                            ),
                            command.id,
                            created_at,
                        ]
                    )
                    session_output_strings.append(
                        tabulate.tabulate(
                            command_table,
                            headers=[
                                f"COMMAND run in {session.name}",
                                "ID",
                                "CREATED",
                            ],
                            tablefmt="plain",
                        )
                    )
            session_output_string = "\n".join(session_output_strings)

            return f"{active_project_title}\n{session_output_string}"

    # TODO (jbai): This requires additional refactoring to decouple network fetching from formatting logic
    def list_ips(
        self, session_name: Optional[str], json_format: bool, all_sessions: bool
    ) -> str:
        project_definition = load_project_or_throw()
        project_id = get_project_id(project_definition.root)

        if all_sessions:
            sessions = get_project_sessions(project_id, session_name, self.api_client)
            sessions = [session.name for session in sessions]
        else:
            sessions = [
                get_project_session(project_id, session_name, self.api_client).name
            ]

        output_json = []
        output_table = []
        for name in sessions:
            head_ip, worker_ips = self._get_head_and_worker_ips(session_name=name)
            if json_format:
                session = get_project_session(project_id, name, self.api_client)
                status = self.get_session_status_and_commands(session=session)[0]
                session_record = {"name": session.name, "status": status}
                project_record = {
                    "name": get_proj_name_from_id(project_id, self.api_client)
                }
                cloud_record = get_cloud_json_from_id(session.cloud_id, self.api_client)
                output_json.append(
                    {
                        "session": session_record,
                        "project": project_record,
                        "cloud": cloud_record,
                        "head-ip": head_ip,
                        "worker-ips": worker_ips,
                    }
                )
            else:
                output_table.append([name, head_ip, "head"])
                for worker_ip in worker_ips:
                    output_table.append([name, worker_ip, "worker"])

        return (
            json.dumps(output_json)
            if json_format
            else tabulate.tabulate(
                output_table,
                headers=["SESSION", "IP ADDRESS", "NODE TYPE"],
                tablefmt="plain",
            )
        )

    # TODO (jbai): Reorganize to other files
    def _get_head_and_worker_ips(self, session_name: str) -> Tuple[str, List[str]]:
        try_import_ray()
        from ray.autoscaler.sdk import get_head_node_ip, get_worker_node_ips

        cluster_config = get_cluster_config(session_name, self.api_client)
        head_ip = get_head_node_ip(cluster_config)
        worker_ips = get_worker_node_ips(cluster_config)

        return head_ip, worker_ips

    # Consolidate this once this https://github.com/anyscale/product/pull/497 gets merged.
    def session_list_json(self, sessions: List[Any], project_id: str) -> str:
        output = []
        for session in sessions:
            status, commands = self.get_session_status_and_commands(session=session)
            record = {"name": session.name}
            record["status"] = status
            record["startup_error"] = (
                session.state_data.startup if session.state_data else {}
            )
            record["stop_error"] = (
                session.state_data.stopping if session.state_data else {}
            )

            record["created_at"] = time.mktime(session.created_at.timetuple())
            record["connect_url"] = session.connect_url
            record["jupyter_notebook_url"] = session.jupyter_notebook_url
            record["ray_dashboard_url"] = session.ray_dashboard_url
            record["grafana_url"] = session.metrics_dashboard_url
            if session.tensorboard_available and session.service_proxy_url:
                record["tensorboard_url"] = session.service_proxy_url
            else:
                record["tensorboard_url"] = None

            record["session_idle_timeout_minutes"] = session.idle_timeout
            record[
                "session_idle_time_remaining_seconds"
            ] = session.idle_time_remaining_seconds

            record["commands"] = commands

            record["project"] = project_id
            record["cloud"] = get_cloud_json_from_id(session.cloud_id, self.api_client)
            del record["cloud"]["config"]
            output.append(record)

        return json.dumps(output)

    def get_session_status_and_commands(
        self, session: Any
    ) -> Tuple[str, List[Dict[str, Any]]]:
        resp = self.api_client.get_session_commands_history_api_v2_session_commands_get(
            session_id=session.id
        )
        commands = []
        is_session_idle = True
        for command in resp.results:
            if command.killed_at is not None:
                command_status = "KILLED"
            elif command.finished_at is not None:
                command_status = "FINISHED"
            else:
                command_status = "RUNNING"
                is_session_idle = False

            command_record = {
                "id": command.id,
                "name": command.name,
                "created_at": humanize_timestamp(command.created_at),
                "status": command_status,
            }
            commands.append(command_record)
        status = str(session.state)
        if status == "Running":
            status = "Idle" if is_session_idle else "TASK_RUNNING"
        return status, commands



# coding: utf-8

"""
    Anyscale API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six

from anyscale_client.configuration import Configuration


class ProductionServiceV2Model(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'id': 'str',
        'name': 'str',
        'description': 'str',
        'project_id': 'str',
        'cloud_id': 'str',
        'creator_id': 'str',
        'created_at': 'datetime',
        'hostname': 'str',
        'current_state': 'ServiceEventCurrentState',
        'goal_state': 'ServiceGoalStates',
        'auth_token': 'str',
        'auto_rollout_enabled': 'bool',
        'versions': 'list[ProductionServiceV2VersionModel]',
        'service_observability_urls': 'ServiceObservabilityUrls'
    }

    attribute_map = {
        'id': 'id',
        'name': 'name',
        'description': 'description',
        'project_id': 'project_id',
        'cloud_id': 'cloud_id',
        'creator_id': 'creator_id',
        'created_at': 'created_at',
        'hostname': 'hostname',
        'current_state': 'current_state',
        'goal_state': 'goal_state',
        'auth_token': 'auth_token',
        'auto_rollout_enabled': 'auto_rollout_enabled',
        'versions': 'versions',
        'service_observability_urls': 'service_observability_urls'
    }

    def __init__(self, id=None, name=None, description=None, project_id=None, cloud_id=None, creator_id=None, created_at=None, hostname=None, current_state=None, goal_state=None, auth_token=None, auto_rollout_enabled=None, versions=None, service_observability_urls=None, local_vars_configuration=None):  # noqa: E501
        """ProductionServiceV2Model - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration()
        self.local_vars_configuration = local_vars_configuration

        self._id = None
        self._name = None
        self._description = None
        self._project_id = None
        self._cloud_id = None
        self._creator_id = None
        self._created_at = None
        self._hostname = None
        self._current_state = None
        self._goal_state = None
        self._auth_token = None
        self._auto_rollout_enabled = None
        self._versions = None
        self._service_observability_urls = None
        self.discriminator = None

        self.id = id
        self.name = name
        if description is not None:
            self.description = description
        self.project_id = project_id
        self.cloud_id = cloud_id
        self.creator_id = creator_id
        self.created_at = created_at
        self.hostname = hostname
        self.current_state = current_state
        self.goal_state = goal_state
        self.auth_token = auth_token
        self.auto_rollout_enabled = auto_rollout_enabled
        self.versions = versions
        self.service_observability_urls = service_observability_urls

    @property
    def id(self):
        """Gets the id of this ProductionServiceV2Model.  # noqa: E501

        Id of the Service  # noqa: E501

        :return: The id of this ProductionServiceV2Model.  # noqa: E501
        :rtype: str
        """
        return self._id

    @id.setter
    def id(self, id):
        """Sets the id of this ProductionServiceV2Model.

        Id of the Service  # noqa: E501

        :param id: The id of this ProductionServiceV2Model.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and id is None:  # noqa: E501
            raise ValueError("Invalid value for `id`, must not be `None`")  # noqa: E501

        self._id = id

    @property
    def name(self):
        """Gets the name of this ProductionServiceV2Model.  # noqa: E501

        Name of the Service  # noqa: E501

        :return: The name of this ProductionServiceV2Model.  # noqa: E501
        :rtype: str
        """
        return self._name

    @name.setter
    def name(self, name):
        """Sets the name of this ProductionServiceV2Model.

        Name of the Service  # noqa: E501

        :param name: The name of this ProductionServiceV2Model.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and name is None:  # noqa: E501
            raise ValueError("Invalid value for `name`, must not be `None`")  # noqa: E501

        self._name = name

    @property
    def description(self):
        """Gets the description of this ProductionServiceV2Model.  # noqa: E501

        Description of the Service  # noqa: E501

        :return: The description of this ProductionServiceV2Model.  # noqa: E501
        :rtype: str
        """
        return self._description

    @description.setter
    def description(self, description):
        """Sets the description of this ProductionServiceV2Model.

        Description of the Service  # noqa: E501

        :param description: The description of this ProductionServiceV2Model.  # noqa: E501
        :type: str
        """

        self._description = description

    @property
    def project_id(self):
        """Gets the project_id of this ProductionServiceV2Model.  # noqa: E501

        Id of the project this Service will start clusters in. This configuration cannot be changed after the service is created.  # noqa: E501

        :return: The project_id of this ProductionServiceV2Model.  # noqa: E501
        :rtype: str
        """
        return self._project_id

    @project_id.setter
    def project_id(self, project_id):
        """Sets the project_id of this ProductionServiceV2Model.

        Id of the project this Service will start clusters in. This configuration cannot be changed after the service is created.  # noqa: E501

        :param project_id: The project_id of this ProductionServiceV2Model.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and project_id is None:  # noqa: E501
            raise ValueError("Invalid value for `project_id`, must not be `None`")  # noqa: E501

        self._project_id = project_id

    @property
    def cloud_id(self):
        """Gets the cloud_id of this ProductionServiceV2Model.  # noqa: E501

        Id of the cloud this Service belongs to, and will launch clusters in. This configuration cannot be changed.  # noqa: E501

        :return: The cloud_id of this ProductionServiceV2Model.  # noqa: E501
        :rtype: str
        """
        return self._cloud_id

    @cloud_id.setter
    def cloud_id(self, cloud_id):
        """Sets the cloud_id of this ProductionServiceV2Model.

        Id of the cloud this Service belongs to, and will launch clusters in. This configuration cannot be changed.  # noqa: E501

        :param cloud_id: The cloud_id of this ProductionServiceV2Model.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and cloud_id is None:  # noqa: E501
            raise ValueError("Invalid value for `cloud_id`, must not be `None`")  # noqa: E501

        self._cloud_id = cloud_id

    @property
    def creator_id(self):
        """Gets the creator_id of this ProductionServiceV2Model.  # noqa: E501

        Id of the user who created the Service  # noqa: E501

        :return: The creator_id of this ProductionServiceV2Model.  # noqa: E501
        :rtype: str
        """
        return self._creator_id

    @creator_id.setter
    def creator_id(self, creator_id):
        """Sets the creator_id of this ProductionServiceV2Model.

        Id of the user who created the Service  # noqa: E501

        :param creator_id: The creator_id of this ProductionServiceV2Model.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and creator_id is None:  # noqa: E501
            raise ValueError("Invalid value for `creator_id`, must not be `None`")  # noqa: E501

        self._creator_id = creator_id

    @property
    def created_at(self):
        """Gets the created_at of this ProductionServiceV2Model.  # noqa: E501

        Time the Service was created  # noqa: E501

        :return: The created_at of this ProductionServiceV2Model.  # noqa: E501
        :rtype: datetime
        """
        return self._created_at

    @created_at.setter
    def created_at(self, created_at):
        """Sets the created_at of this ProductionServiceV2Model.

        Time the Service was created  # noqa: E501

        :param created_at: The created_at of this ProductionServiceV2Model.  # noqa: E501
        :type: datetime
        """
        if self.local_vars_configuration.client_side_validation and created_at is None:  # noqa: E501
            raise ValueError("Invalid value for `created_at`, must not be `None`")  # noqa: E501

        self._created_at = created_at

    @property
    def hostname(self):
        """Gets the hostname of this ProductionServiceV2Model.  # noqa: E501

        The hostname of the service  # noqa: E501

        :return: The hostname of this ProductionServiceV2Model.  # noqa: E501
        :rtype: str
        """
        return self._hostname

    @hostname.setter
    def hostname(self, hostname):
        """Sets the hostname of this ProductionServiceV2Model.

        The hostname of the service  # noqa: E501

        :param hostname: The hostname of this ProductionServiceV2Model.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and hostname is None:  # noqa: E501
            raise ValueError("Invalid value for `hostname`, must not be `None`")  # noqa: E501

        self._hostname = hostname

    @property
    def current_state(self):
        """Gets the current_state of this ProductionServiceV2Model.  # noqa: E501

        The current state of this service  # noqa: E501

        :return: The current_state of this ProductionServiceV2Model.  # noqa: E501
        :rtype: ServiceEventCurrentState
        """
        return self._current_state

    @current_state.setter
    def current_state(self, current_state):
        """Sets the current_state of this ProductionServiceV2Model.

        The current state of this service  # noqa: E501

        :param current_state: The current_state of this ProductionServiceV2Model.  # noqa: E501
        :type: ServiceEventCurrentState
        """
        if self.local_vars_configuration.client_side_validation and current_state is None:  # noqa: E501
            raise ValueError("Invalid value for `current_state`, must not be `None`")  # noqa: E501

        self._current_state = current_state

    @property
    def goal_state(self):
        """Gets the goal_state of this ProductionServiceV2Model.  # noqa: E501

        The goal state of this service  # noqa: E501

        :return: The goal_state of this ProductionServiceV2Model.  # noqa: E501
        :rtype: ServiceGoalStates
        """
        return self._goal_state

    @goal_state.setter
    def goal_state(self, goal_state):
        """Sets the goal_state of this ProductionServiceV2Model.

        The goal state of this service  # noqa: E501

        :param goal_state: The goal_state of this ProductionServiceV2Model.  # noqa: E501
        :type: ServiceGoalStates
        """
        if self.local_vars_configuration.client_side_validation and goal_state is None:  # noqa: E501
            raise ValueError("Invalid value for `goal_state`, must not be `None`")  # noqa: E501

        self._goal_state = goal_state

    @property
    def auth_token(self):
        """Gets the auth_token of this ProductionServiceV2Model.  # noqa: E501

        Token to use for service auth. To use the token, add it as a header with the key 'Authorization' and the value 'Bearer <token>'  # noqa: E501

        :return: The auth_token of this ProductionServiceV2Model.  # noqa: E501
        :rtype: str
        """
        return self._auth_token

    @auth_token.setter
    def auth_token(self, auth_token):
        """Sets the auth_token of this ProductionServiceV2Model.

        Token to use for service auth. To use the token, add it as a header with the key 'Authorization' and the value 'Bearer <token>'  # noqa: E501

        :param auth_token: The auth_token of this ProductionServiceV2Model.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and auth_token is None:  # noqa: E501
            raise ValueError("Invalid value for `auth_token`, must not be `None`")  # noqa: E501

        self._auth_token = auth_token

    @property
    def auto_rollout_enabled(self):
        """Gets the auto_rollout_enabled of this ProductionServiceV2Model.  # noqa: E501

        Whether or not the service is using auto rollout  # noqa: E501

        :return: The auto_rollout_enabled of this ProductionServiceV2Model.  # noqa: E501
        :rtype: bool
        """
        return self._auto_rollout_enabled

    @auto_rollout_enabled.setter
    def auto_rollout_enabled(self, auto_rollout_enabled):
        """Sets the auto_rollout_enabled of this ProductionServiceV2Model.

        Whether or not the service is using auto rollout  # noqa: E501

        :param auto_rollout_enabled: The auto_rollout_enabled of this ProductionServiceV2Model.  # noqa: E501
        :type: bool
        """
        if self.local_vars_configuration.client_side_validation and auto_rollout_enabled is None:  # noqa: E501
            raise ValueError("Invalid value for `auto_rollout_enabled`, must not be `None`")  # noqa: E501

        self._auto_rollout_enabled = auto_rollout_enabled

    @property
    def versions(self):
        """Gets the versions of this ProductionServiceV2Model.  # noqa: E501

        Active versions of this service, sorted by creation time in ascending order.  # noqa: E501

        :return: The versions of this ProductionServiceV2Model.  # noqa: E501
        :rtype: list[ProductionServiceV2VersionModel]
        """
        return self._versions

    @versions.setter
    def versions(self, versions):
        """Sets the versions of this ProductionServiceV2Model.

        Active versions of this service, sorted by creation time in ascending order.  # noqa: E501

        :param versions: The versions of this ProductionServiceV2Model.  # noqa: E501
        :type: list[ProductionServiceV2VersionModel]
        """
        if self.local_vars_configuration.client_side_validation and versions is None:  # noqa: E501
            raise ValueError("Invalid value for `versions`, must not be `None`")  # noqa: E501

        self._versions = versions

    @property
    def service_observability_urls(self):
        """Gets the service_observability_urls of this ProductionServiceV2Model.  # noqa: E501

        A JSON object with useful urls pointing to Grafana dashboards.  # noqa: E501

        :return: The service_observability_urls of this ProductionServiceV2Model.  # noqa: E501
        :rtype: ServiceObservabilityUrls
        """
        return self._service_observability_urls

    @service_observability_urls.setter
    def service_observability_urls(self, service_observability_urls):
        """Sets the service_observability_urls of this ProductionServiceV2Model.

        A JSON object with useful urls pointing to Grafana dashboards.  # noqa: E501

        :param service_observability_urls: The service_observability_urls of this ProductionServiceV2Model.  # noqa: E501
        :type: ServiceObservabilityUrls
        """
        if self.local_vars_configuration.client_side_validation and service_observability_urls is None:  # noqa: E501
            raise ValueError("Invalid value for `service_observability_urls`, must not be `None`")  # noqa: E501

        self._service_observability_urls = service_observability_urls

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, ProductionServiceV2Model):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, ProductionServiceV2Model):
            return True

        return self.to_dict() != other.to_dict()



import json
import os
from os import path
import shlex
import subprocess
from typing import Any, Dict, List, Optional, Sequence, Union

import click
import yaml

from anyscale.cli_logger import BlockLogger
from anyscale.client.openapi_client import (
    AppConfig,
    Build,
    ComputeTemplateConfig,
    CreateComputeTemplate,
    Session,
    StopSessionOptions,
)
from anyscale.cluster_config import get_cluster_config
from anyscale.controllers.base_controller import BaseController
from anyscale.project import (
    get_and_validate_project_id,
    get_project_id,
    get_project_session,
    get_project_sessions,
    load_project_or_throw,
)
from anyscale.sdk.anyscale_client import StartSessionOptions, UpdateSession
from anyscale.sdk.anyscale_client.models.create_cluster import CreateCluster
from anyscale.shared_anyscale_utils.util import get_container_name, slugify
from anyscale.util import (
    canonicalize_remote_location,
    get_endpoint,
    get_working_dir,
    wait_for_session_start,
)
from anyscale.utils.aws_credentials_util import (
    get_credentials_as_env_vars_from_cluster_config,
)
from anyscale.utils.env_utils import set_env
from anyscale.utils.imports.all import try_import_ray
from anyscale.utils.name_utils import gen_valid_name


def get_head_node_ip(cluster_config: Union[Dict[str, Any], str]) -> Any:
    try_import_ray()
    from ray.autoscaler.sdk import get_head_node_ip

    return get_head_node_ip(cluster_config)


def rsync(*args: Any, **kwargs: Any) -> None:
    try_import_ray()
    from ray.autoscaler.sdk import rsync as ray_rsync

    ray_rsync(*args, **kwargs)


class SessionController(BaseController):
    def __init__(
        self, log: Optional[BlockLogger] = None, initialize_auth_api_client: bool = True
    ):
        if log is None:
            log = BlockLogger()

        super().__init__(initialize_auth_api_client=initialize_auth_api_client)
        self.log = log

    def stop(
        self,
        session_name: Optional[str],
        delete: bool,
        workers_only: bool,
        keep_min_workers: bool,
    ) -> None:
        project_definition = load_project_or_throw()
        project_id = get_project_id(project_definition.root)
        sessions = get_project_sessions(
            project_id, session_name, self.api_client, all_active_states=True
        )

        if not session_name and len(sessions) > 1:
            raise click.ClickException(
                "Multiple active sessions: {}\n"
                "Please specify the one you want to stop with --cluster-name.".format(
                    [session.name for session in sessions]
                )
            )

        for session in sessions:
            # Stop the session and mark it as stopped in the database.
            self.api_client.stop_session_api_v2_sessions_session_id_stop_post(
                session.id,
                StopSessionOptions(
                    terminate=True,
                    workers_only=workers_only,
                    keep_min_workers=keep_min_workers,
                    delete=delete,
                ),
            )

        session_names = [session.name for session in sessions]
        session_names_str = ", ".join(session_names)
        url = get_endpoint(f"/projects/{project_id}")
        self.log.info(
            f"Session {session_names_str} shutting down. View progress at {url}"
        )

    def start(
        self,
        session_name: Optional[str],
        build_id: Optional[str],
        build_identifier: Optional[str],
        compute_config: str,
        idle_timeout: Optional[int],
        sync_files: bool,
    ) -> None:
        if sync_files:
            raise click.ClickException(
                "--sync-files has been deprecated and `anyscale start` will no longer ensure files "
                "synced to the working directory of the head node will also be synced on the workers."
            )
        message = (
            "Warning: `anyscale start` has been deprecated. Please try "
            'using `ray.init("anyscale://cluster_name")` for your use case and let the Anyscale team '
            "know if you are having any issues."
        )
        self.log.warning(message)

        project_definition = load_project_or_throw()
        project_id = get_project_id(project_definition.root)
        session_name = self._get_or_generate_session_name(session_name, project_id)

        assert (
            build_id or build_identifier
        ), "--build-id or --build-identifier must be specified."
        if not build_id and build_identifier:
            build_id = self._get_build_id_from_build_identifier(
                build_identifier, project_id
            )
        assert build_id  # for mypy

        compute_template_id = self._register_compute_config(compute_config, project_id)

        session = self._create_or_update_session_data(
            session_name, project_id, build_id, compute_template_id, idle_timeout
        )
        self.anyscale_api_client.start_session(
            session.id,
            StartSessionOptions(
                build_id=build_id, compute_template_id=compute_template_id
            ),
        )

        wait_for_session_start(project_id, session_name, self.api_client)
        url = get_endpoint(f"/projects/{project_id}/clusters/{session.id}")
        print(f"Session {session_name} finished starting. View at {url}")

    def ssh(  # noqa: PLR0912, PLR0913
        self,
        session_name: str,
        ssh_option: Sequence[str],
        worker_node_id: Optional[str],
        worker_node_ip: Optional[str],
        force_head_node: bool,
        project_id: Optional[str],
        cmd: Optional[str] = None,
    ) -> None:

        config_path = "/home/ray/ray_bootstrap_config.yaml"
        on_head_node = path.exists(config_path)

        ssh_to_head = worker_node_id is None and worker_node_ip is None

        if force_head_node and ssh_to_head:
            raise click.ClickException(
                "Can't force head node if not sshing to a worker_node."
            )

        if worker_node_id is not None and worker_node_ip is not None:
            raise click.ClickException(
                "A worker node id and worker node ip cannot both be set at once."
            )

        if force_head_node or (on_head_node and (worker_node_id or worker_node_ip)):
            # On the head node
            print("SSHing from head node to worker node.")

            # Import ray where it's needed here to avoid making ray a
            # regular dependency of the CLI. This should only execute on
            # a head node where we can be confident ray will already
            # be installed.
            ray = try_import_ray()

            with open(config_path) as f:
                config_dict = yaml.safe_load(f)

            try:
                ssh_user = config_dict["auth"]["ssh_user"]
                ssh_key_path = config_dict["auth"]["ssh_private_key"]
                container_name = config_dict["docker"]["container_name"]
            except ValueError as e:
                raise click.ClickException(
                    f"Could not parse ~/ray_bootstrap_config.yaml: {e}"
                )

            if worker_node_ip is not None:
                node_ip = worker_node_ip
            elif worker_node_id is not None:
                ray.init(address="auto")
                nodes = ray.nodes()
                node_info = next(
                    (node for node in nodes if node.get("NodeID") == worker_node_id),
                    None,
                )
                if node_info is None:
                    raise click.ClickException(
                        f"Could not find a worker node with id {worker_node_id}."
                    )
                node_ip = node_info.get("NodeManagerAddress")

            ssh_command = (
                ["ssh"]
                + list(ssh_option)
                + ["-tt", "-i", ssh_key_path]
                + [f"{ssh_user}@{node_ip}"]
                + [f"docker exec -it {container_name} sh -c 'which bash && bash || sh'"]
            )
        else:
            # On a laptop
            project_id = get_and_validate_project_id(
                project_id=project_id,
                project_name=None,
                parent_cloud_id=None,
                api_client=self.api_client,
                anyscale_api_client=self.anyscale_api_client,
            )
            session = get_project_session(project_id, session_name, self.api_client)

            cluster_config = get_cluster_config(
                session_name,
                self.api_client,
                disable_project_sync=True,
                project_id=project_id,
            )

            head_ip = self.api_client.get_session_head_ip_api_v2_sessions_session_id_head_ip_get(
                session.id
            ).result.head_ip

            ssh_user = cluster_config["auth"]["ssh_user"]
            key_path = cluster_config["auth"]["ssh_private_key"]
            container_name = get_container_name(cluster_config)

            base_ssh_command = (
                ["ssh"]
                + list(ssh_option)
                + ["-tt", "-i", key_path]
                + [f"{ssh_user}@{head_ip}"]
            )
            if cmd:
                command = shlex.quote(cmd)
            else:
                command = shlex.quote("which bash && bash || sh")

            if ssh_to_head:
                ssh_command = base_ssh_command + (
                    [f"docker exec -it {container_name} sh -c {command}"]
                    if container_name
                    else []
                )
            else:
                if not container_name:
                    raise click.ClickException(
                        "Cannot ssh to a worker node if this cluster has no container."
                    )
                # Don't use the ssh_options for the second ssh command because
                # they probably refer to files that exist on the laptop but not
                # the head node.

                if worker_node_id is not None:
                    ssh_flag = f"--worker-node-id {worker_node_id}"
                else:
                    ssh_flag = f"--worker-node-ip {worker_node_ip}"
                ssh_from_head_command = (
                    f"'anyscale ssh {ssh_flag} --force-head-node True'"
                )
                ssh_command = base_ssh_command + (
                    [f"docker exec -it {container_name} sh -c  {ssh_from_head_command}"]
                )

        subprocess.run(ssh_command)

    def pull(
        self,
        session_name: Optional[str] = None,
        source: Optional[str] = None,
        target: Optional[str] = None,
        config: Optional[str] = None,
        rsync_exclude_override: Optional[List[str]] = None,
    ) -> None:
        project_definition = load_project_or_throw()

        try:
            self.log.info("Collecting files from remote.")
            project_id = get_project_id(project_definition.root)
            cluster_config = get_cluster_config(session_name, self.api_client)
            if rsync_exclude_override is not None:
                cluster_config["rsync_exclude"] = rsync_exclude_override
            directory_name = get_working_dir(
                cluster_config, project_id, self.api_client
            )
            source_directory = f"{directory_name}/"

            aws_credentials = get_credentials_as_env_vars_from_cluster_config(
                cluster_config
            )

            source = canonicalize_remote_location(cluster_config, source, project_id)

            with set_env(**aws_credentials):
                if source and target:
                    target = os.path.expanduser(target)
                    rsync(
                        cluster_config, source=source, target=target, down=True,
                    )
                elif source or target:
                    raise click.ClickException(
                        "Source and target are not both specified. Please either specify both or neither."
                    )
                else:
                    rsync(
                        cluster_config,
                        source=source_directory,
                        target=project_definition.root,
                        down=True,
                    )

            if config:
                session = get_project_session(project_id, session_name, self.api_client)
                resp = self.api_client.get_session_cluster_config_api_v2_sessions_session_id_cluster_config_get(
                    session.id
                )
                cluster_config = yaml.safe_load(resp.result.config_with_defaults)
                with open(config, "w") as f:
                    yaml.dump(cluster_config, f, default_flow_style=False)

            self.log.info("Pull completed.")

        except Exception as e:  # noqa: BLE001
            print(e)
            raise click.ClickException(str(e))  # type: ignore

    def push(
        self,
        session_name: Optional[str],
        source: Optional[str],
        target: Optional[str],
        config: Optional[str],  # noqa: ARG002
        all_nodes: bool,
        rsync_exclude_override: Optional[List[str]] = None,
    ) -> None:
        if all_nodes:
            raise click.ClickException(
                "The `all_nodes` option is deprecated and will be removed in the future. "
                "Pushing to worker nodes is not reliable since workers may be "
                "added during autoscaling. Going forward, Anyscale will only support syncing "
                "working directory files to the head and workers."
            )

        project_definition = load_project_or_throw()
        project_id = get_project_id(project_definition.root)
        session = get_project_session(project_id, session_name, self.api_client)
        session_name = session.name

        cluster_config = get_cluster_config(session_name, self.api_client)
        if rsync_exclude_override is not None:
            cluster_config["rsync_exclude"] = rsync_exclude_override
        target = canonicalize_remote_location(cluster_config, target, project_id)

        aws_credentials = get_credentials_as_env_vars_from_cluster_config(
            cluster_config
        )

        with set_env(**aws_credentials):
            if source and target:
                rsync(
                    cluster_config, source=source, target=target, down=False,
                )
            elif source or target:
                raise click.ClickException(
                    "Source and target are not both specified. Please either specify both or neither."
                )
            else:
                rsync(
                    cluster_config, source=None, target=None, down=False,
                )

        url = get_endpoint(f"/projects/{project_id}/clusters/{session.id}")
        self.log.info(f"Pushed to session {session_name}. View at {url}")

    # Helpers

    def _get_or_generate_session_name(
        self, session_name: Optional[str], project_id: Optional[str]
    ) -> str:
        """
        Return slugified session name if provided, else generate default session name from project id
        """
        assert (
            session_name or project_id
        ), "Either the session_name or project_id must be provided"

        if not session_name:
            session_name = str(
                self.api_client.get_project_default_session_name_api_v2_projects_project_id_default_session_name_get(
                    project_id=project_id,
                ).result.name
            )
        else:
            session_name = slugify(session_name)
        assert session_name
        return session_name

    def _register_compute_config(
        self, compute_config_file: str, project_id: str
    ) -> str:
        """
        Register the compute config with Anyscale and get its ID.
        """
        with open(compute_config_file) as f:
            config_dict = json.load(f)
        config_object = ComputeTemplateConfig(**config_dict)
        created_template = self.api_client.create_compute_template_api_v2_compute_templates_post(
            create_compute_template=CreateComputeTemplate(
                name=gen_valid_name("autogenerated-config"),
                project_id=project_id,
                config=config_object,
            )
        ).result
        compute_template_id = str(created_template.id)
        return compute_template_id

    def _create_or_update_session_data(
        self,
        session_name: str,
        project_id: str,
        build_id: str,
        compute_template_id: str,
        idle_timeout: Optional[int],
    ) -> Session:
        """
        Creates new session with app configs if session with `session_name` doesn't
        already exist. Otherwise, update the `idle_timeout` of the existing
        session if provided.
        """
        session_list = self.api_client.list_sessions_api_v2_sessions_get(
            project_id=project_id, active_only=False, name=session_name
        ).results

        session_exists = len(session_list) > 0
        if not session_exists:
            # Create a new session if there is no existing session with the givens session_name
            create_session_data = CreateCluster(
                name=session_name,
                project_id=project_id,
                cluster_environment_build_id=build_id,
                cluster_compute_id=compute_template_id,
                idle_timeout_minutes=idle_timeout,
            )
            session = self.anyscale_api_client.create_cluster(
                create_session_data
            ).result
        else:
            # Get the existing session and update the idle_timeout if required
            session = session_list[0]
            if idle_timeout:
                self.anyscale_api_client.update_session(
                    session.id, UpdateSession(idle_timeout=idle_timeout)
                )

        return session

    def _get_build_id_from_build_identifier(
        self, build_identifier: str, project_id: str
    ) -> str:
        # TODO(nikita): App config API should support getting a build ID from the build identifier.
        components = build_identifier.rsplit(":", 1)
        app_config_name = components[0]
        # try catch this
        try:
            app_config_revision = int(components[1]) if len(components) > 1 else None
        except ValueError:
            raise click.ClickException(
                "Invalid build-identifier provided. Please make sure the build identifier is of "
                "the form <app-config-name>:<build-revision>."
            )

        app_config_id = self._get_app_config_id_from_name(app_config_name, project_id)

        builds = self._list_builds(app_config_id)
        if app_config_revision:
            for build in builds:
                if build.revision == app_config_revision:
                    return str(build.id)

            raise click.ClickException(
                "Revision {} of app config '{}' not found.".format(
                    app_config_revision, app_config_name
                )
            )
        else:
            latest_build_revision = -1
            build_to_use = None
            for build in builds:
                if build.revision > latest_build_revision:
                    latest_build_revision = build.revision
                    build_to_use = build

            if not build_to_use:
                raise click.ClickException(
                    "Error finding latest build of app config {}. Please manually "
                    "specify the build version in the build identifier.".format(
                        app_config_name
                    )
                )
            return str(build_to_use.id)

    def _get_app_config_id_from_name(
        self, app_config_name: str, project_id: str
    ) -> str:
        app_configs = self._list_app_configs(project_id)
        for app_config in app_configs:
            if app_config.name == app_config_name:
                return str(app_config.id)

        raise click.ClickException(
            f"Application config '{app_config_name}' not found. "
            + "Available app configs: {}".format(", ".join(a.name for a in app_configs))
        )

    def _list_app_configs(self, project_id: str) -> List[AppConfig]:
        entities = []
        has_more = True
        while has_more:
            resp = self.anyscale_api_client.list_app_configs(
                project_id=project_id, count=50, paging_token=None
            )
            entities.extend(resp.results)
            paging_token = resp.metadata.next_paging_token
            has_more = paging_token is not None
        return entities

    def _list_builds(self, app_config_id: str) -> List[Build]:
        entities = []
        has_more = True
        while has_more:
            resp = self.anyscale_api_client.list_builds(
                app_config_id, count=50, paging_token=None
            )
            entities.extend(resp.results)
            paging_token = resp.metadata.next_paging_token
            has_more = paging_token is not None
        return entities

    def _resolve_session(
        self, session_name: str, project_name: Optional[str] = None
    ) -> Session:
        """
        Resolves a session by name.
        This is distinct from  `anyscale.project.get_project_session` because:
        1. we rely on project names instead of ids
        2. we allow non-active sessions to be resolved

        Raises an exception if the session does not exist.

        Params
        session_name - name of the session
        project_name - optional project name that the session is in;
                       if absent, we use the workspace's project
        """
        if project_name:
            projects = self.api_client.list_projects_api_v2_projects_get().results
            project = next(
                project for project in projects if project.name == project_name
            )
            project_id = project.id
        else:
            project_definition = load_project_or_throw()
            project_id = get_project_id(project_definition.root)

        sessions_list = self.api_client.list_sessions_api_v2_sessions_get(
            project_id, name=session_name
        ).results

        if len(sessions_list) == 0:
            raise click.ClickException(
                f"No session found with name {session_name} in project {project_id}"
            )

        return sessions_list[0]



# coding: utf-8

"""
    Managed Ray API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six

from openapi_client.configuration import Configuration


class MiniJobRun(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'id': 'str',
        'ray_session_name': 'str',
        'ray_job_id': 'str',
        'name': 'str',
        'status': 'BaseJobStatus',
        'created_at': 'datetime',
        'finished_at': 'datetime',
        'ha_job_id': 'str',
        'ray_job_submission_id': 'str',
        'cluster_id': 'str',
        'namespace_id': 'str',
        'environment_id': 'str',
        'project_id': 'str',
        'creator_id': 'str',
        'integration_execution_details_id': 'str',
        'bucket_log_prefix': 'str',
        'cluster': 'MiniCluster',
        'integration_details': 'IntegrationDetails'
    }

    attribute_map = {
        'id': 'id',
        'ray_session_name': 'ray_session_name',
        'ray_job_id': 'ray_job_id',
        'name': 'name',
        'status': 'status',
        'created_at': 'created_at',
        'finished_at': 'finished_at',
        'ha_job_id': 'ha_job_id',
        'ray_job_submission_id': 'ray_job_submission_id',
        'cluster_id': 'cluster_id',
        'namespace_id': 'namespace_id',
        'environment_id': 'environment_id',
        'project_id': 'project_id',
        'creator_id': 'creator_id',
        'integration_execution_details_id': 'integration_execution_details_id',
        'bucket_log_prefix': 'bucket_log_prefix',
        'cluster': 'cluster',
        'integration_details': 'integration_details'
    }

    def __init__(self, id=None, ray_session_name=None, ray_job_id=None, name=None, status=None, created_at=None, finished_at=None, ha_job_id=None, ray_job_submission_id=None, cluster_id=None, namespace_id='DEPRECATED_NAMESPACE_ID', environment_id=None, project_id=None, creator_id=None, integration_execution_details_id=None, bucket_log_prefix=None, cluster=None, integration_details=None, local_vars_configuration=None):  # noqa: E501
        """MiniJobRun - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration()
        self.local_vars_configuration = local_vars_configuration

        self._id = None
        self._ray_session_name = None
        self._ray_job_id = None
        self._name = None
        self._status = None
        self._created_at = None
        self._finished_at = None
        self._ha_job_id = None
        self._ray_job_submission_id = None
        self._cluster_id = None
        self._namespace_id = None
        self._environment_id = None
        self._project_id = None
        self._creator_id = None
        self._integration_execution_details_id = None
        self._bucket_log_prefix = None
        self._cluster = None
        self._integration_details = None
        self.discriminator = None

        self.id = id
        self.ray_session_name = ray_session_name
        self.ray_job_id = ray_job_id
        if name is not None:
            self.name = name
        self.status = status
        self.created_at = created_at
        if finished_at is not None:
            self.finished_at = finished_at
        if ha_job_id is not None:
            self.ha_job_id = ha_job_id
        if ray_job_submission_id is not None:
            self.ray_job_submission_id = ray_job_submission_id
        self.cluster_id = cluster_id
        if namespace_id is not None:
            self.namespace_id = namespace_id
        self.environment_id = environment_id
        if project_id is not None:
            self.project_id = project_id
        self.creator_id = creator_id
        if integration_execution_details_id is not None:
            self.integration_execution_details_id = integration_execution_details_id
        if bucket_log_prefix is not None:
            self.bucket_log_prefix = bucket_log_prefix
        self.cluster = cluster
        if integration_details is not None:
            self.integration_details = integration_details

    @property
    def id(self):
        """Gets the id of this MiniJobRun.  # noqa: E501


        :return: The id of this MiniJobRun.  # noqa: E501
        :rtype: str
        """
        return self._id

    @id.setter
    def id(self, id):
        """Sets the id of this MiniJobRun.


        :param id: The id of this MiniJobRun.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and id is None:  # noqa: E501
            raise ValueError("Invalid value for `id`, must not be `None`")  # noqa: E501

        self._id = id

    @property
    def ray_session_name(self):
        """Gets the ray_session_name of this MiniJobRun.  # noqa: E501


        :return: The ray_session_name of this MiniJobRun.  # noqa: E501
        :rtype: str
        """
        return self._ray_session_name

    @ray_session_name.setter
    def ray_session_name(self, ray_session_name):
        """Sets the ray_session_name of this MiniJobRun.


        :param ray_session_name: The ray_session_name of this MiniJobRun.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and ray_session_name is None:  # noqa: E501
            raise ValueError("Invalid value for `ray_session_name`, must not be `None`")  # noqa: E501

        self._ray_session_name = ray_session_name

    @property
    def ray_job_id(self):
        """Gets the ray_job_id of this MiniJobRun.  # noqa: E501


        :return: The ray_job_id of this MiniJobRun.  # noqa: E501
        :rtype: str
        """
        return self._ray_job_id

    @ray_job_id.setter
    def ray_job_id(self, ray_job_id):
        """Sets the ray_job_id of this MiniJobRun.


        :param ray_job_id: The ray_job_id of this MiniJobRun.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and ray_job_id is None:  # noqa: E501
            raise ValueError("Invalid value for `ray_job_id`, must not be `None`")  # noqa: E501

        self._ray_job_id = ray_job_id

    @property
    def name(self):
        """Gets the name of this MiniJobRun.  # noqa: E501


        :return: The name of this MiniJobRun.  # noqa: E501
        :rtype: str
        """
        return self._name

    @name.setter
    def name(self, name):
        """Sets the name of this MiniJobRun.


        :param name: The name of this MiniJobRun.  # noqa: E501
        :type: str
        """

        self._name = name

    @property
    def status(self):
        """Gets the status of this MiniJobRun.  # noqa: E501


        :return: The status of this MiniJobRun.  # noqa: E501
        :rtype: BaseJobStatus
        """
        return self._status

    @status.setter
    def status(self, status):
        """Sets the status of this MiniJobRun.


        :param status: The status of this MiniJobRun.  # noqa: E501
        :type: BaseJobStatus
        """
        if self.local_vars_configuration.client_side_validation and status is None:  # noqa: E501
            raise ValueError("Invalid value for `status`, must not be `None`")  # noqa: E501

        self._status = status

    @property
    def created_at(self):
        """Gets the created_at of this MiniJobRun.  # noqa: E501


        :return: The created_at of this MiniJobRun.  # noqa: E501
        :rtype: datetime
        """
        return self._created_at

    @created_at.setter
    def created_at(self, created_at):
        """Sets the created_at of this MiniJobRun.


        :param created_at: The created_at of this MiniJobRun.  # noqa: E501
        :type: datetime
        """
        if self.local_vars_configuration.client_side_validation and created_at is None:  # noqa: E501
            raise ValueError("Invalid value for `created_at`, must not be `None`")  # noqa: E501

        self._created_at = created_at

    @property
    def finished_at(self):
        """Gets the finished_at of this MiniJobRun.  # noqa: E501


        :return: The finished_at of this MiniJobRun.  # noqa: E501
        :rtype: datetime
        """
        return self._finished_at

    @finished_at.setter
    def finished_at(self, finished_at):
        """Sets the finished_at of this MiniJobRun.


        :param finished_at: The finished_at of this MiniJobRun.  # noqa: E501
        :type: datetime
        """

        self._finished_at = finished_at

    @property
    def ha_job_id(self):
        """Gets the ha_job_id of this MiniJobRun.  # noqa: E501


        :return: The ha_job_id of this MiniJobRun.  # noqa: E501
        :rtype: str
        """
        return self._ha_job_id

    @ha_job_id.setter
    def ha_job_id(self, ha_job_id):
        """Sets the ha_job_id of this MiniJobRun.


        :param ha_job_id: The ha_job_id of this MiniJobRun.  # noqa: E501
        :type: str
        """

        self._ha_job_id = ha_job_id

    @property
    def ray_job_submission_id(self):
        """Gets the ray_job_submission_id of this MiniJobRun.  # noqa: E501


        :return: The ray_job_submission_id of this MiniJobRun.  # noqa: E501
        :rtype: str
        """
        return self._ray_job_submission_id

    @ray_job_submission_id.setter
    def ray_job_submission_id(self, ray_job_submission_id):
        """Sets the ray_job_submission_id of this MiniJobRun.


        :param ray_job_submission_id: The ray_job_submission_id of this MiniJobRun.  # noqa: E501
        :type: str
        """

        self._ray_job_submission_id = ray_job_submission_id

    @property
    def cluster_id(self):
        """Gets the cluster_id of this MiniJobRun.  # noqa: E501


        :return: The cluster_id of this MiniJobRun.  # noqa: E501
        :rtype: str
        """
        return self._cluster_id

    @cluster_id.setter
    def cluster_id(self, cluster_id):
        """Sets the cluster_id of this MiniJobRun.


        :param cluster_id: The cluster_id of this MiniJobRun.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and cluster_id is None:  # noqa: E501
            raise ValueError("Invalid value for `cluster_id`, must not be `None`")  # noqa: E501

        self._cluster_id = cluster_id

    @property
    def namespace_id(self):
        """Gets the namespace_id of this MiniJobRun.  # noqa: E501

        ID of the Anyscale Namespace.  # noqa: E501

        :return: The namespace_id of this MiniJobRun.  # noqa: E501
        :rtype: str
        """
        return self._namespace_id

    @namespace_id.setter
    def namespace_id(self, namespace_id):
        """Sets the namespace_id of this MiniJobRun.

        ID of the Anyscale Namespace.  # noqa: E501

        :param namespace_id: The namespace_id of this MiniJobRun.  # noqa: E501
        :type: str
        """

        self._namespace_id = namespace_id

    @property
    def environment_id(self):
        """Gets the environment_id of this MiniJobRun.  # noqa: E501


        :return: The environment_id of this MiniJobRun.  # noqa: E501
        :rtype: str
        """
        return self._environment_id

    @environment_id.setter
    def environment_id(self, environment_id):
        """Sets the environment_id of this MiniJobRun.


        :param environment_id: The environment_id of this MiniJobRun.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and environment_id is None:  # noqa: E501
            raise ValueError("Invalid value for `environment_id`, must not be `None`")  # noqa: E501

        self._environment_id = environment_id

    @property
    def project_id(self):
        """Gets the project_id of this MiniJobRun.  # noqa: E501


        :return: The project_id of this MiniJobRun.  # noqa: E501
        :rtype: str
        """
        return self._project_id

    @project_id.setter
    def project_id(self, project_id):
        """Sets the project_id of this MiniJobRun.


        :param project_id: The project_id of this MiniJobRun.  # noqa: E501
        :type: str
        """

        self._project_id = project_id

    @property
    def creator_id(self):
        """Gets the creator_id of this MiniJobRun.  # noqa: E501


        :return: The creator_id of this MiniJobRun.  # noqa: E501
        :rtype: str
        """
        return self._creator_id

    @creator_id.setter
    def creator_id(self, creator_id):
        """Sets the creator_id of this MiniJobRun.


        :param creator_id: The creator_id of this MiniJobRun.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and creator_id is None:  # noqa: E501
            raise ValueError("Invalid value for `creator_id`, must not be `None`")  # noqa: E501

        self._creator_id = creator_id

    @property
    def integration_execution_details_id(self):
        """Gets the integration_execution_details_id of this MiniJobRun.  # noqa: E501


        :return: The integration_execution_details_id of this MiniJobRun.  # noqa: E501
        :rtype: str
        """
        return self._integration_execution_details_id

    @integration_execution_details_id.setter
    def integration_execution_details_id(self, integration_execution_details_id):
        """Sets the integration_execution_details_id of this MiniJobRun.


        :param integration_execution_details_id: The integration_execution_details_id of this MiniJobRun.  # noqa: E501
        :type: str
        """

        self._integration_execution_details_id = integration_execution_details_id

    @property
    def bucket_log_prefix(self):
        """Gets the bucket_log_prefix of this MiniJobRun.  # noqa: E501


        :return: The bucket_log_prefix of this MiniJobRun.  # noqa: E501
        :rtype: str
        """
        return self._bucket_log_prefix

    @bucket_log_prefix.setter
    def bucket_log_prefix(self, bucket_log_prefix):
        """Sets the bucket_log_prefix of this MiniJobRun.


        :param bucket_log_prefix: The bucket_log_prefix of this MiniJobRun.  # noqa: E501
        :type: str
        """

        self._bucket_log_prefix = bucket_log_prefix

    @property
    def cluster(self):
        """Gets the cluster of this MiniJobRun.  # noqa: E501


        :return: The cluster of this MiniJobRun.  # noqa: E501
        :rtype: MiniCluster
        """
        return self._cluster

    @cluster.setter
    def cluster(self, cluster):
        """Sets the cluster of this MiniJobRun.


        :param cluster: The cluster of this MiniJobRun.  # noqa: E501
        :type: MiniCluster
        """
        if self.local_vars_configuration.client_side_validation and cluster is None:  # noqa: E501
            raise ValueError("Invalid value for `cluster`, must not be `None`")  # noqa: E501

        self._cluster = cluster

    @property
    def integration_details(self):
        """Gets the integration_details of this MiniJobRun.  # noqa: E501


        :return: The integration_details of this MiniJobRun.  # noqa: E501
        :rtype: IntegrationDetails
        """
        return self._integration_details

    @integration_details.setter
    def integration_details(self, integration_details):
        """Sets the integration_details of this MiniJobRun.


        :param integration_details: The integration_details of this MiniJobRun.  # noqa: E501
        :type: IntegrationDetails
        """

        self._integration_details = integration_details

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, MiniJobRun):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, MiniJobRun):
            return True

        return self.to_dict() != other.to_dict()



import sys
from typing import List, Optional

import click
from typing_extensions import Literal

from anyscale.cli_logger import BlockLogger
from anyscale.controllers.service_controller import ServiceController
from anyscale.util import validate_non_negative_arg


log = BlockLogger()  # CLI Logger


@click.group(
    "service", help="Interact with production services running on Anyscale.",
)
def service_cli() -> None:
    pass


@service_cli.command(name="deploy", help="Deploy a service to Anyscale.")
@click.argument("service-config-file", required=True)
@click.option("--name", "-n", required=False, default=None, help="Name of service.")
@click.option(
    "--description", required=False, default=None, help="Description of service."
)
@click.option(
    "--healthcheck-url", required=False, help="Healthcheck URL for Service.",
)
@click.argument(
    "entrypoint", required=False, nargs=-1,
)
def deploy(
    service_config_file: str,
    entrypoint: List[str],
    name: Optional[str],
    description: Optional[str],
    healthcheck_url: Optional[str],
) -> None:
    # TODO[Bruce]: Remove update flag once fully test apply function.
    """Deploy a service (create or update).

    This is *only* supported for v1 services.
    """
    service_controller = ServiceController()
    service_controller.deploy(
        service_config_file,
        name=name,
        description=description,
        healthcheck_url=healthcheck_url,
        entrypoint=entrypoint,
        is_entrypoint_cmd="--" in sys.argv,
    )


@service_cli.command(
    name="rollout",
    help="Rollout a Service v2 to Anyscale. Please contact support for details.",
)
@click.option(
    "-f",
    "--service-config-file",
    required=True,
    help="The path of the service configuration file.",
)
@click.option("--name", "-n", required=False, default=None, help="Name of service.")
@click.option("--version", required=False, default=None, help="Version of service.")
@click.option(
    "--canary-percent",
    required=False,
    default=None,
    type=int,
    help="The percentage of traffic going to the new version of Service that you are deploying",
)
@click.option(
    "--rollout-strategy",
    required=False,
    default=None,
    type=click.Choice(["ROLLOUT", "IN_PLACE"], case_sensitive=False),
    help="Strategy for rollout.",
)
def rollout(
    service_config_file: str,
    name: Optional[str],
    version: Optional[str],
    canary_percent: Optional[int],
    rollout_strategy: Optional[Literal["ROLLOUT", "IN_PLACE"]],
) -> None:
    """Start or update a service rollout to a new version.

    This is *only* supported for v2 services.
    """
    service_controller = ServiceController()
    service_controller.rollout(
        service_config_file,
        name=name,
        version=version,
        canary_percent=canary_percent,
        rollout_strategy=rollout_strategy,
    )


@service_cli.command(name="list", help="Display information about existing services.")
@click.option(
    "--name", "-n", required=False, default=None, help="Filter by service name."
)
@click.option(
    "--service-id", "--id", required=False, default=None, help="Filter by service id."
)
@click.option(
    "--project-id", required=False, default=None, help="Filter by project id."
)
@click.option(
    "--created-by-me",
    help="List services created by me only.",
    is_flag=True,
    default=False,
)
@click.option(
    "--max-items",
    required=False,
    default=10,
    type=int,
    help="Max items to show in list.",
    callback=validate_non_negative_arg,
)
def list(  # noqa: A001
    name: Optional[str],
    service_id: Optional[str],
    project_id: Optional[str],
    created_by_me: bool,
    max_items: int,
) -> None:
    """List services based on the provided filters.

    This returns both v1 and v2 services.
    """
    service_controller = ServiceController()
    service_controller.list(
        name=name,
        service_id=service_id,
        project_id=project_id,
        created_by_me=created_by_me,
        max_items=max_items,
    )


@service_cli.command(name="archive", help="Archive a service.")
@click.option("--service-id", "--id", required=False, help="ID of service.")
@click.option("--name", "-n", required=False, help="Name of service.")
@click.option("--project-id", required=False, help="Filter by project id.")
@click.option(
    "--service-config-file", "-f", help="The path of the service configuration file.",
)
def archive(
    service_id: Optional[str],
    name: Optional[str],
    project_id: Optional[str],
    service_config_file: Optional[str],
) -> None:
    """Archive a service, which must already be terminated.

    This is currently only supported for v1 services but should be extended to v2.
    """
    service_controller = ServiceController()
    service_id = service_controller.get_service_id(
        service_id=service_id,
        service_name=name,
        service_config_file=service_config_file,
        project_id=project_id,
    )
    service_controller.archive(service_id)


@service_cli.command(
    name="rollback", help="Attempt to rollback a v2 service asynchronously."
)
@click.option(
    "--service-id", "--id", default=None, help="ID of service.",
)
@click.option("--name", "-n", required=False, default=None, help="Name of service.")
@click.option("--project-id", required=False, help="Filter by project id.")
@click.option(
    "--service-config-file", "-f", help="The path of the service configuration file.",
)
def rollback(
    service_id: Optional[str],
    name: Optional[str],
    project_id: Optional[str],
    service_config_file: Optional[str],
) -> None:
    """Perform a rollback for a service that is currently in a rollout.

    This *only* applies to v2 services.
    """
    service_controller = ServiceController()
    service_id = service_controller.get_service_id(
        service_id=service_id,
        service_name=name,
        service_config_file=service_config_file,
        project_id=project_id,
    )
    service_controller.rollback(service_id)


@service_cli.command(
    name="terminate", help="Attempt to terminate a service asynchronously."
)
@click.option(
    "--service-id", "--id", required=False, help="ID of service.",
)
@click.option("--name", "-n", required=False, help="Name of service.")
@click.option("--project-id", required=False, help="Filter by project id.")
@click.option(
    "--service-config-file", "-f", help="The path of the service configuration file.",
)
def terminate(
    service_id: Optional[str],
    name: Optional[str],
    project_id: Optional[str],
    service_config_file: Optional[str],
) -> None:
    """Terminate a service.

    This applies to both v1 and v2 services.
    """
    service_controller = ServiceController()
    service_id = service_controller.get_service_id(
        service_id=service_id,
        service_name=name,
        service_config_file=service_config_file,
        project_id=project_id,
    )
    service_controller.terminate(service_id)



from typing import Any, Dict, List, Optional
from unittest.mock import _Call, ANY, call, Mock, patch

import pytest

from anyscale.connect_utils.prepare_cluster import (
    PrepareClusterAction,
    PrepareClusterBlock,
)
from anyscale.sdk.anyscale_client import (
    CreateCluster,
    StartClusterOptions,
    UpdateCluster,
)


@pytest.fixture()
def mock_prepare_cluster_block():
    with patch.multiple(
        "anyscale.connect_utils.prepare_cluster.PrepareClusterBlock",
        __init__=Mock(return_value=None),
    ):
        prepare_cluster_block = PrepareClusterBlock()
        prepare_cluster_block.anyscale_api_client = Mock()
        prepare_cluster_block.api_client = Mock()
        prepare_cluster_block._ray = Mock()
        prepare_cluster_block.log = Mock()
        prepare_cluster_block.block_label = ""
        return prepare_cluster_block


@pytest.mark.parametrize("cluster_name", [None, "test_cluster_name"])
@pytest.mark.parametrize("ray_cli_is_connected", [True, False])
def test_start_or_create_cluster(
    mock_prepare_cluster_block, cluster_name: Optional[str], ray_cli_is_connected: bool
):
    mock_prepare_cluster_block.anyscale_api_client.list_sessions = Mock(
        return_value=Mock(results=[], metadata=Mock(next_paging_token=None))
    )
    mock_prepare_cluster_block._ray.util.client.ray.is_connected = Mock(
        return_value=ray_cli_is_connected
    )
    mock_cluster = Mock()
    mock_prepare_cluster_block._create_or_update_session_data = Mock(
        return_value=(mock_cluster, True)
    )
    mock_prepare_cluster_block._start_cluster_if_required = Mock()

    mock_prepare_cluster_block._start_or_create_cluster(
        "test_project_id",
        "test_build_id",
        "test_compute_template_id",
        cluster_name,
        100,
        None,
    )

    if not cluster_name:
        mock_prepare_cluster_block.anyscale_api_client.list_sessions.assert_called_once_with(
            "test_project_id", count=50, paging_token=None
        )
    if ray_cli_is_connected:
        mock_prepare_cluster_block._ray.util.disconnect.assert_called_once_with()
    mock_prepare_cluster_block._create_or_update_session_data.assert_called_once_with(
        cluster_name if cluster_name else "cluster-0",
        "test_project_id",
        "test_build_id",
        "test_compute_template_id",
        100,
        False,
        not cluster_name,
    )
    mock_prepare_cluster_block._start_cluster_if_required.assert_called_once_with(
        mock_cluster,
        True,
        "test_project_id",
        "test_build_id",
        "test_compute_template_id",
        None,
    )


@pytest.mark.parametrize("start_required", [True, False])
@pytest.mark.parametrize("allow_public_internet_traffic", [True, False])
def test_start_cluster_if_required(
    mock_prepare_cluster_block,
    start_required: bool,
    allow_public_internet_traffic: bool,
):
    mock_cluster = Mock(id="test_cluster_id")
    mock_cluster.name = "test_cluster_name"

    mock_prepare_cluster_block._create_or_update_session_data = Mock(
        return_value=(mock_cluster, start_required)
    )
    mock_prepare_cluster_block._log_cluster_configs = Mock()
    mock_wait_for_session_start = Mock()
    with patch.multiple(
        "anyscale.connect_utils.prepare_cluster",
        wait_for_session_start=mock_wait_for_session_start,
    ):
        mock_prepare_cluster_block._start_cluster_if_required(
            mock_cluster,
            start_required,
            "test_project_id",
            "test_build_id",
            "test_compute_template_id",
            allow_public_internet_traffic,
        )
    mock_prepare_cluster_block._log_cluster_configs.assert_called_once_with(
        mock_cluster, "test_build_id", "test_compute_template_id", ANY
    )
    if start_required:
        mock_prepare_cluster_block.anyscale_api_client.start_cluster.assert_called_once_with(
            mock_cluster.id,
            StartClusterOptions(
                cluster_environment_build_id="test_build_id",
                cluster_compute_id="test_compute_template_id",
                allow_public_internet_traffic=allow_public_internet_traffic,
            ),
        )
        mock_wait_for_session_start.assert_called_once_with(
            "test_project_id",
            "test_cluster_name",
            mock_prepare_cluster_block.api_client,
            log=mock_prepare_cluster_block.log,
            block_label=mock_prepare_cluster_block.block_label,
        )


@pytest.mark.parametrize("cluster_exists", [True, False])
@pytest.mark.parametrize("start_required", [True, False])
@pytest.mark.parametrize("idle_timeout", [None, 100])
def test_create_or_update_session_data(
    mock_prepare_cluster_block,
    cluster_exists: bool,
    start_required: bool,
    idle_timeout: Optional[int],
):
    mock_cluster = Mock(state="Running", id="test_cluster_id")
    mock_get_cluster = Mock(
        side_effect=[(mock_cluster if cluster_exists else None), mock_cluster]
    )
    mock_prepare_cluster_block._validate_new_cluster_compute_and_env_match_existing_cluster = Mock(
        return_value=start_required
    )

    with patch.multiple(
        "anyscale.connect_utils.prepare_cluster", get_cluster=mock_get_cluster
    ):
        mock_prepare_cluster_block._create_or_update_session_data(
            "test_cluster_name",
            "test_project_id",
            "test_build_id",
            "test_compute_template_id",
            idle_timeout,
            None,
        )

    if not cluster_exists:
        mock_get_cluster.assert_called_with(
            mock_prepare_cluster_block.anyscale_api_client,
            "test_project_id",
            "test_cluster_name",
        )
        assert mock_get_cluster.call_count == 2
        mock_prepare_cluster_block.anyscale_api_client.create_cluster.assert_called_once_with(
            CreateCluster(
                name="test_cluster_name",
                project_id="test_project_id",
                cluster_environment_build_id="test_build_id",
                cluster_compute_id="test_compute_template_id",
                idle_timeout_minutes=idle_timeout,
                allow_public_internet_traffic=None,
            )
        )
        mock_prepare_cluster_block.anyscale_api_client.update_cluster.assert_not_called()
        mock_prepare_cluster_block._validate_new_cluster_compute_and_env_match_existing_cluster.assert_not_called()
    else:
        mock_get_cluster.assert_called_once_with(
            mock_prepare_cluster_block.anyscale_api_client,
            "test_project_id",
            "test_cluster_name",
        )
        mock_prepare_cluster_block.anyscale_api_client.create_cluster.assert_not_called()
        mock_prepare_cluster_block._validate_new_cluster_compute_and_env_match_existing_cluster.assert_called_once_with(
            "test_project_id", mock_cluster, print_warnings=False
        )
        if idle_timeout:
            mock_prepare_cluster_block.anyscale_api_client.update_cluster.assert_called_once_with(
                mock_cluster.id, UpdateCluster(idle_timeout_minutes=idle_timeout)
            )


def test_create_or_update_session_data_generates_new_cluster_name(
    mock_prepare_cluster_block,
):
    mock_taken_cluster = Mock(
        state="Running", id="test_cluster_id", name="cluster-exists"
    )
    mock_new_cluster = Mock(
        state="Running", id="test_cluster_id", name="cluster-newgen"
    )

    mock_get_cluster = Mock(side_effect=[mock_taken_cluster, None, mock_new_cluster])
    mock_prepare_cluster_block._validate_new_cluster_compute_and_env_match_existing_cluster = (
        Mock()
    )

    with patch.multiple(
        "anyscale.connect_utils.prepare_cluster", get_cluster=mock_get_cluster
    ):
        mock_prepare_cluster_block._create_or_update_session_data(
            "cluster-exists",
            "test_project_id",
            "test_build_id",
            "test_compute_template_id",
            100,
            None,
            True,
        )

    mock_get_cluster.assert_has_calls(
        [
            call(
                mock_prepare_cluster_block.anyscale_api_client,
                "test_project_id",
                "cluster-exists",
            ),
            call(
                mock_prepare_cluster_block.anyscale_api_client, "test_project_id", ANY,
            ),
            call(
                mock_prepare_cluster_block.anyscale_api_client, "test_project_id", ANY,
            ),
        ]
    )
    # assert the newly generated cluster has a different name than the existing cluster
    cluster_name_of_first_get_cluster_call = mock_get_cluster.call_args_list[0][0][-1]
    cluster_name_of_last_get_cluster_call = mock_get_cluster.call_args_list[-1][0][-1]
    assert (
        cluster_name_of_first_get_cluster_call != cluster_name_of_last_get_cluster_call
    )

    mock_prepare_cluster_block.anyscale_api_client.create_cluster.assert_called_once_with(
        CreateCluster(
            name=ANY,
            project_id="test_project_id",
            cluster_environment_build_id="test_build_id",
            cluster_compute_id="test_compute_template_id",
            idle_timeout_minutes=100,
            allow_public_internet_traffic=None,
        )
    )
    mock_prepare_cluster_block.anyscale_api_client.update_cluster.assert_not_called()
    mock_prepare_cluster_block._validate_new_cluster_compute_and_env_match_existing_cluster.assert_not_called()


@pytest.mark.parametrize("needs_update", [True, False])
@pytest.mark.parametrize("cluster_exists", [True, False])
@pytest.mark.parametrize("cluster_name", ["test_cluster_name", None])
@pytest.mark.parametrize("cluster_state", ["Running", "Other"])
def test_get_prepare_cluster_action(
    mock_prepare_cluster_block,
    needs_update: bool,
    cluster_exists: bool,
    cluster_name: Optional[str],
    cluster_state: str,
):
    mock_cluster = Mock(state=cluster_state)
    mock_cluster.name = cluster_name
    mock_prepare_cluster_block.anyscale_api_client.search_clusters = Mock(
        return_value=Mock(results=[mock_cluster] if cluster_exists else [])
    )
    mock_prepare_cluster_block.anyscale_api_client.get_session = Mock(
        return_value=Mock(result=mock_cluster if cluster_exists else None)
    )

    prepare_cluster_action = mock_prepare_cluster_block._get_prepare_cluster_action(
        "test_project_id", cluster_name, needs_update
    )
    if not cluster_name or not cluster_exists:
        assert prepare_cluster_action == PrepareClusterAction.CREATE
    elif cluster_state != "Running":
        assert prepare_cluster_action == PrepareClusterAction.START
    elif needs_update:
        assert prepare_cluster_action == PrepareClusterAction.UPDATE
    else:
        assert prepare_cluster_action == PrepareClusterAction.NO_OP

    if cluster_name:
        mock_prepare_cluster_block.anyscale_api_client.search_clusters.assert_called_once_with(
            {
                "project_id": "test_project_id",
                "name": {"equals": cluster_name},
                "archive_status": "ALL",
            },
        )


@pytest.mark.parametrize("cluster_env_name", ["test_cluster_env_name", None])
def test_get_cluster_build(mock_prepare_cluster_block, cluster_env_name: Optional[str]):
    mock_build = Mock()
    mock_prepare_cluster_block._get_cluster_env_build = Mock(return_value=mock_build)
    mock_prepare_cluster_block._get_default_cluster_env_build = Mock(
        return_value=mock_build
    )

    assert (
        mock_prepare_cluster_block._get_cluster_build(cluster_env_name, None)
        == mock_build
    )

    if cluster_env_name:
        mock_prepare_cluster_block._get_cluster_env_build.assert_called_once_with(
            cluster_env_name, None
        )
    else:
        mock_prepare_cluster_block._get_default_cluster_env_build.assert_called_once_with()


@pytest.mark.parametrize("clust_env_revision", [None, 1])
def test_get_cluster_env_build(
    mock_prepare_cluster_block, clust_env_revision: Optional[int]
):
    mock_cluster_env = Mock(id="test_app_template_id")
    mock_cluster_env.name = "test_cluster_env_name"
    mock_prepare_cluster_block.anyscale_api_client.list_app_configs = Mock(
        return_value=Mock(
            results=[mock_cluster_env], metadata=Mock(next_paging_token=None)
        )
    )
    mock_build1 = Mock(id="build1", revision=1)
    mock_build2 = Mock(id="build2", revision=2)
    mock_prepare_cluster_block.anyscale_api_client.list_builds = Mock(
        return_value=Mock(
            results=[mock_build1, mock_build2], metadata=Mock(next_paging_token=None)
        )
    )

    if clust_env_revision == 1:
        assert (
            mock_prepare_cluster_block._get_cluster_env_build(
                "test_cluster_env_name", clust_env_revision
            )
            == mock_build1
        )
    else:
        assert (
            mock_prepare_cluster_block._get_cluster_env_build(
                "test_cluster_env_name", clust_env_revision
            )
            == mock_build2
        )


@pytest.mark.parametrize("build_pr", [None, "test_build_pr"])
@pytest.mark.parametrize("build_commit", [None, "test_build_commit"])
@pytest.mark.parametrize("cluster_env_name", [None, "test_cluster_env:name"])
@pytest.mark.parametrize("cluster_env_dict", [None, {"key": "val"}])
def test_build_cluster_env_if_needed(
    mock_prepare_cluster_block,
    build_pr: Optional[str],
    build_commit: Optional[str],
    cluster_env_name: Optional[str],
    cluster_env_dict: Optional[Dict[str, Any]],
):
    mock_prepare_cluster_block._build_app_config_from_source = Mock(
        return_value="test_built_cluster_env_name"
    )

    observed_result = mock_prepare_cluster_block._build_cluster_env_if_needed(
        "test_project_id",
        build_pr,
        build_commit,
        cluster_env_dict,
        cluster_env_name,
        False,
    )
    if build_pr or build_commit:
        assert observed_result == "test_built_cluster_env_name"
    elif cluster_env_dict:
        if cluster_env_name:
            assert observed_result == "test_cluster_env-name"
        else:
            assert observed_result.startswith("anonymous_cluster_env-")
        mock_prepare_cluster_block.anyscale_api_client.create_app_config.assert_called_once_with(
            {
                "name": observed_result,
                "project_id": "test_project_id",
                "config_json": cluster_env_dict,
            }
        )
    else:
        assert observed_result == cluster_env_name


@pytest.mark.parametrize("cluster_compute_name", [None, "test_cluster_compute_name"])
@pytest.mark.parametrize("cluster_compute_dict", [None, {"cloud_id": "mock_cloud_id"}])
@pytest.mark.parametrize("cloud_name", [None, "test_cloud_name"])
def test_get_cluster_compute_id(
    mock_prepare_cluster_block,
    cluster_compute_name: Optional[str],
    cluster_compute_dict: Optional[Dict[str, str]],
    cloud_name: Optional[str],
):
    cloud_id = "test_cloud_id"
    mock_get_selected_cloud_id_or_default = Mock(return_value=cloud_id)
    mock_default_config_obj = Mock()
    mock_config_obj_from_cluster_compute_dict = Mock()
    mock_prepare_cluster_block.anyscale_api_client.get_default_compute_config = Mock(
        return_value=Mock(result=mock_default_config_obj)
    )
    mock_prepare_cluster_block._register_compute_template = Mock(
        return_value="mock_registered_template"
    )
    mock_prepare_cluster_block._get_cluster_compute_id_from_name = Mock(
        return_value="mock_cluster_compute_template"
    )

    with patch.multiple(
        "anyscale.connect_utils.prepare_cluster",
        ComputeTemplateConfig=Mock(
            return_value=mock_config_obj_from_cluster_compute_dict
        ),
        get_selected_cloud_id_or_default=mock_get_selected_cloud_id_or_default,
    ):
        observed_result = mock_prepare_cluster_block._get_cluster_compute_id(
            "test_project_id", cluster_compute_name, cluster_compute_dict, cloud_name
        )

    if cluster_compute_name:
        mock_prepare_cluster_block._get_cluster_compute_id_from_name.assert_called_once_with(
            "test_project_id", cluster_compute_name
        )
        assert observed_result == "mock_cluster_compute_template"
    else:
        if cluster_compute_dict:
            mock_config_obj = mock_config_obj_from_cluster_compute_dict
        else:
            mock_prepare_cluster_block.anyscale_api_client.get_default_compute_config.assert_called_once_with(
                cloud_id
            )
            mock_config_obj = mock_default_config_obj
        mock_prepare_cluster_block._register_compute_template.assert_called_once_with(
            "test_project_id", mock_config_obj
        )
        assert observed_result == "mock_registered_template"


@pytest.mark.parametrize("cluster_computes", [[], [Mock(id="test_cluster_compute_id")]])
def test_get_cluster_compute_id_from_name(
    mock_prepare_cluster_block, cluster_computes: List[Any]
):
    mock_prepare_cluster_block.api_client.search_compute_templates_api_v2_compute_templates_search_post = Mock(
        return_value=Mock(results=cluster_computes)
    )
    if len(cluster_computes) == 0:
        with pytest.raises(ValueError):
            mock_prepare_cluster_block._get_cluster_compute_id_from_name(
                "test_project_id", "test_cluster_compute_name"
            )
    else:
        assert (
            mock_prepare_cluster_block._get_cluster_compute_id_from_name(
                "test_project_id", "test_cluster_compute_name"
            )
            == cluster_computes[0].id
        )


def test_is_equal_cluster_compute(mock_prepare_cluster_block):
    def mock_get_compute_template(cluster_compute_id: str):
        if cluster_compute_id == "cluster_compute_1_id":
            return Mock(result=Mock(config=cluster_compute_1))
        elif cluster_compute_id == "cluster_compute_2_id":
            return Mock(result=Mock(config=cluster_compute_2))

    mock_prepare_cluster_block.anyscale_api_client.get_compute_template = Mock(
        side_effect=mock_get_compute_template
    )

    # Test cluster computes are equal
    cluster_compute_1 = "test_cluster_compute"
    cluster_compute_2 = "test_cluster_compute"
    assert mock_prepare_cluster_block._is_equal_cluster_compute(
        "cluster_compute_1_id", "cluster_compute_2_id"
    )

    # Test cluster_computes are different
    cluster_compute_1 = "test_cluster_compute"
    cluster_compute_2 = "test_diff_cluster_compute"
    assert not mock_prepare_cluster_block._is_equal_cluster_compute(
        "cluster_compute_1_id", "cluster_compute_2_id"
    )


@pytest.mark.parametrize("build_matches", [True, False])
@pytest.mark.parametrize("cluster_compute_matches", [True, False])
@pytest.mark.parametrize("allow_public_internet_traffic_matches", [True, False])
def test_validate_new_cluster_compute_and_env_match_existing_cluster(
    mock_prepare_cluster_block,
    build_matches: bool,
    cluster_compute_matches: bool,
    allow_public_internet_traffic_matches: bool,
):
    mock_cluster = Mock()
    mock_cluster.build_id = "mock_build_id1"
    mock_cluster.allow_public_internet_traffic = True
    mock_prepare_cluster_block.cluster_env_name = "test_cluster_env_name"
    mock_prepare_cluster_block.cluster_env_revision = "test_cluster_env_revision"
    mock_prepare_cluster_block.cluster_compute_name = "test_cluster_compute_name"
    mock_prepare_cluster_block.cluster_compute_dict = "test_cluster_compute_dict"
    mock_prepare_cluster_block.cloud_name = "test_cloud_name"
    mock_prepare_cluster_block.cluster_env_dict = "test_cluster_env_dict"
    mock_prepare_cluster_block._get_cluster_compute_id = Mock()

    if build_matches:
        mock_prepare_cluster_block._get_cluster_build = Mock(
            return_value=Mock(id="mock_build_id1")
        )
    else:
        mock_prepare_cluster_block._get_cluster_build = Mock(
            return_value=Mock(id="mock_build_id2")
        )

    mock_prepare_cluster_block._is_equal_cluster_compute = Mock(
        return_value=cluster_compute_matches
    )

    mock_prepare_cluster_block.allow_public_internet_traffic = (
        allow_public_internet_traffic_matches
    )

    observed_result = mock_prepare_cluster_block._validate_new_cluster_compute_and_env_match_existing_cluster(
        "test_projet_id", mock_cluster,
    )
    if (
        not build_matches
        or cluster_compute_matches
        or not allow_public_internet_traffic_matches
    ):
        assert observed_result


@pytest.mark.parametrize(
    "get_prepare_cluster_action",
    [
        PrepareClusterAction.CREATE,
        PrepareClusterAction.START,
        PrepareClusterAction.NO_OP,
    ],
)
@pytest.mark.parametrize("cluster_compute_name", ["test_cluster_compute_name", None])
@pytest.mark.parametrize("cluster_env_name", ["test_cluster_env_name", None])
def test_init(
    get_prepare_cluster_action: bool,
    cluster_compute_name: Optional[str],
    cluster_env_name: Optional[str],
):
    cluster_name = "test_cluster_name"
    mock_get_prepare_cluster_action = Mock(return_value=get_prepare_cluster_action)
    mock_cluster = Mock(
        cloud_id="mock_cloud_id", compute_template_id="mock_cluster_compute_id"
    )
    mock_cluster.name = "test_cluster_name"
    mock_build_cluster_env_if_needed = Mock(return_value=cluster_env_name)
    mock_get_cluster_build = Mock(return_value=Mock(id="mock_build_id"))
    mock_get_selected_cloud_id_or_default = Mock(return_value="mock_cloud_id")
    mock_get_cluster_compute_id = Mock(return_value="mock_cluster_compute_id")
    mock_wait_for_app_build = Mock()
    mock_start_or_create_cluster = Mock(return_value="test_cluster_name")
    mock_validate_new_cluster_compute_and_env_match_existing_cluster = Mock()
    mock_log_cluster_configs = Mock()
    mock_cluster_compute_dict = Mock()
    mock_force_rebuild = Mock()
    mock_cluster_env_dict = Mock()
    mock_cluster_env_revision = Mock()

    with patch.multiple(
        "anyscale.connect_utils.prepare_cluster.PrepareClusterBlock",
        _log_cluster_configs=mock_log_cluster_configs,
        _validate_new_cluster_compute_and_env_match_existing_cluster=mock_validate_new_cluster_compute_and_env_match_existing_cluster,
        _start_or_create_cluster=mock_start_or_create_cluster,
        _wait_for_app_build=mock_wait_for_app_build,
        _get_cluster_compute_id=mock_get_cluster_compute_id,
        _get_cluster_build=mock_get_cluster_build,
        _build_cluster_env_if_needed=mock_build_cluster_env_if_needed,
        _get_prepare_cluster_action=mock_get_prepare_cluster_action,
    ), patch.multiple(
        "anyscale.connect_utils.prepare_cluster",
        get_cluster=Mock(return_value=mock_cluster),
        get_auth_api_client=Mock(return_value=Mock()),
        get_selected_cloud_id_or_default=mock_get_selected_cloud_id_or_default,
    ):
        prepare_cluster_block = PrepareClusterBlock(
            project_id="test_project_id",
            cluster_name=cluster_name,
            autosuspend_timeout=10,
            allow_public_internet_traffic=None,
            needs_update=False,
            cluster_compute_name=cluster_compute_name,
            cluster_compute_dict=mock_cluster_compute_dict,
            cloud_name="test_cloud_name",
            build_pr=None,
            force_rebuild=mock_force_rebuild,
            build_commit=None,
            cluster_env_name=cluster_env_name,
            cluster_env_dict=mock_cluster_env_dict,
            cluster_env_revision=mock_cluster_env_revision,
            ray=Mock(),
        )
        prepare_cluster_block.prepare()
        assert prepare_cluster_block.cluster_name == "test_cluster_name"

    mock_get_prepare_cluster_action.assert_called_once_with(
        "test_project_id", cluster_name, False
    )
    if get_prepare_cluster_action in {
        PrepareClusterAction.CREATE,
        PrepareClusterAction.START,
    }:
        mock_build_cluster_env_if_needed.assert_called_once_with(
            "test_project_id",
            None,
            None,
            mock_cluster_env_dict,
            cluster_env_name,
            mock_force_rebuild,
        )
        if cluster_env_name or not cluster_name:
            mock_get_cluster_build.assert_called_once_with(
                cluster_env_name, mock_cluster_env_revision
            )
            build_id = "mock_build_id"
        else:
            build_id = mock_cluster.build_id

        if cluster_compute_name:
            mock_get_cluster_compute_id.assert_called_once_with(
                "test_project_id",
                cluster_compute_name,
                mock_cluster_compute_dict,
                "test_cloud_name",
            )
        mock_wait_for_app_build.assert_called_once_with("test_project_id", build_id)
        mock_start_or_create_cluster.assert_called_once_with(
            project_id="test_project_id",
            build_id=build_id,
            compute_template_id="mock_cluster_compute_id",
            cluster_name=cluster_name,
            autosuspend_timeout=10,
            allow_public_internet_traffic=None,
        )
    else:
        mock_validate_new_cluster_compute_and_env_match_existing_cluster.assert_called_once_with(
            project_id="test_project_id", running_cluster=mock_cluster
        )
        mock_log_cluster_configs.assert_called_once_with(
            mock_cluster, mock_cluster.build_id, mock_cluster.compute_template_id, ANY,
        )
        mock_start_or_create_cluster.assert_not_called()


@pytest.mark.parametrize("idle_timeout", [60, -1])
@pytest.mark.parametrize("maximum_uptime_minutes", [60, -1])
def test_log_cluster_configs(
    mock_prepare_cluster_block, idle_timeout: int, maximum_uptime_minutes: int,
) -> None:
    mock_prepare_cluster_block.anyscale_api_client.get_build = Mock(
        return_value=Mock(result=Mock(application_template_id="env_name", revision=1))
    )
    mock_compute_config_config = Mock(
        id="id", config=Mock(maximum_uptime_minutes=maximum_uptime_minutes)
    )
    mock_compute_config_config.name = "config_name"
    mock_prepare_cluster_block.anyscale_api_client.get_compute_template = Mock(
        return_value=Mock(result=mock_compute_config_config)
    )
    mock_cluster = Mock(id="ses_1", idle_timeout=idle_timeout)
    mock_build_id = "build_id"
    mock_compute_template_id = "compute_template_id"
    mock_url = "url"
    mock_prepare_cluster_block._log_cluster_configs(
        mock_cluster, mock_build_id, mock_compute_template_id, mock_url
    )
    mock_prepare_cluster_block.anyscale_api_client.get_build.assert_called_once_with(
        mock_build_id
    )
    mock_prepare_cluster_block.anyscale_api_client.get_compute_template.assert_called_once_with(
        mock_compute_template_id
    )

    def call_args_for_log_line(name: str, value: str) -> _Call:
        return call(f"{' ' * 2}{(name + ':').ljust(30)}{value}", block_label="")

    def get_auto_terminate_output(value: int) -> str:
        return f"{value} minutes" if value > 0 else "disabled"

    mock_prepare_cluster_block.log.info.assert_has_calls(
        [
            call_args_for_log_line("cluster id", "ses_1"),
            call_args_for_log_line("cluster environment", "env_name:1"),
            call_args_for_log_line("cluster environment id", "build_id"),
            call_args_for_log_line("cluster compute", "config_name"),
            call_args_for_log_line("cluster compute id", "compute_template_id"),
            call_args_for_log_line(
                "idle termination", get_auto_terminate_output(idle_timeout)
            ),
            call_args_for_log_line(
                "maximum uptime", get_auto_terminate_output(maximum_uptime_minutes)
            ),
            call_args_for_log_line("link", "url"),
        ]
    )
